import re

import pandas as pd
from collections import defaultdict
import itertools

def sort_chr_list( chr_list ):
  sort_chr_list = [chrom for chrom in chr_list if re.match( r'chr\d+$', chrom)]
  sort_chr_list.sort(key=lambda arr: (arr[:3], int(arr[3:])))
  if 'chrM' in chr_list:  sort_chr_list.append('chrM')
  if 'chrMT' in chr_list:  sort_chr_list.append('chrMT')
  if 'chrX' in chr_list:  sort_chr_list.append('chrX')
  if 'chrY' in chr_list:  sort_chr_list.append('chrY')
  return sort_chr_list



def get_gene_df( gtf ):
  feature_list = []
  for ln in open(gtf):
    if ln.startswith('#'): continue
    fields = ln.split('\t')
    if fields[2] != 'gene': continue
    chrom, start, end, strand = fields[0], int( fields[3]), int( fields[4]), fields[6]
    attr_field = fields[-1]
    if not chrom.startswith('chr'): chrom = 'chr' + chrom

    # if 'gene_name "' not in attr_field: continue
    # if 'gene_biotype "' not in attr_field: continue
    for keywords in [ ['gene_type "', 'gene_name "'],  ['gene_biotype "', 'gene_name "'],  ['gene_biotype "', 'gene "']]:
      try:
        gene_name, genetype = attr_field.split(keywords[1])[1].split('"')[0], attr_field.split(keywords[0])[1].split('"')[0]
        feature_list.append( [chrom, start, end, strand, gene_name, genetype] )
        break
      except:
        pass

  df = pd.DataFrame(feature_list, columns =["chrom", "start", "end", "strand", "gene_name", "genetype"] )
  return df

# df  = get_gene_df( r'H:\tmpDownload\gencode.v38.annotation.gtf\gencode.v38.annotation.gtf' )

# from collections import Counter
# print( df.shape )
# print( Counter( df['biotype'] ) )


def to_ranges(iterable):
    iterable = sorted(set(iterable))
    for key, group in itertools.groupby(enumerate(iterable),
                                        lambda t: t[1] - t[0]):
        group = list(group)
        yield group[0][1], group[-1][1]

def get_gene_merge_exon_dic(gtf):
  dd = defaultdict( list )
  gene_chr_strand = {}
  for ln in open(gtf):
    if ln.startswith('#'): continue
    if 'gene_biotype "protein_coding"' not in ln: continue
    fields = ln.split('\t')
    if fields[2] != 'exon': continue
    start, end = int( fields[3] ), int( fields[4] )
    try:
      gene_name = fields[-1].split('gene_name "')[1].split('"')[0]
    except:
      continue
    dd[gene_name].append(  range(start, end)  )
    chrom, strand = fields[0], fields[6]
    if not chrom.startswith('chr'): chrom = 'chr'+ chrom
    gene_chr_strand[gene_name] = [chrom, strand]

  gene_exon_range_dic = {}
  for gene, range_list in dd.items():
    l = []
    for r in range_list:
      l.extend( list(r) )
    range_list = list( to_ranges(l) )
    if len( range_list) > 0:
      gene_exon_range_dic[gene] = {'chrom':gene_chr_strand[gene][0], 'strand':gene_chr_strand[gene][1], 'exon':range_list  }
  return gene_exon_range_dic

# d = get_gene_merge_exon_dic(r'../ensembl_Homo_sapiens.GRCh38.93.gtf')
# for k in list( d.keys() )[:5]:
#   print(k, d[k] )

def gene_df2dic(df):
  dic = {'+': {}, '-': {}}
  gene_name_list = list( df['gene_name'] )
  chr_list = list( df['chrom'] )
  start_list = list( df['start'] )
  end_list = list( df['end'] )
  strand_list = list( df['strand'] )
  for i, gene in enumerate(gene_name_list):
    strand = strand_list[i]
    dic[strand][gene] = [chr_list[i], start_list[i], end_list[i]]
  return dic


import os, sys, json, shutil, subprocess, multiprocessing
from shutil import which

def get_file_size( filePath ):
  fsize = os.path.getsize(filePath)
  fsize = fsize/float(1024*1024)
  return str(round(fsize,2) ) + 'Mb'

def wc_count(file_name):
  out = subprocess.getoutput("wc -l %s" % file_name)
  return int(out.split()[0])

def json2dic(json_dir):
  with open(json_dir, 'r') as f:
    dict = json.load(fp=f)
  return dict

def dic2json(dict, json_dir):
  with open(json_dir, 'w') as f:
    json.dump(dict, f)

def dic2csv(dict, csv_dir):
  with open(csv_dir, 'w') as f:
    for k, v in dict.items():
      f.write( str(k) + ',' + str(v) + '\n')

def two_list_2_dict(key_list, value_list):
  return dict(zip(key_list, value_list))

def read_tmp_variable(name=None):
  # 写死内部数据
  if not os.path.exists('./tmp_variable.json'):
    dic2json({}, './tmp_variable.json' )
  if name:
    return json2dic('./tmp_variable.json')[name]
  else:
    return json2dic('./tmp_variable.json')

def write_tmp_variable(name, value):
  tmp_variable_dic = read_tmp_variable()
  tmp_variable_dic[name] = value
  dic2json(tmp_variable_dic, './tmp_variable.json')

def write_string_to_file(string,  file):
  with open(file, 'w') as f:
    f.write(string)

def create_output_file( output_dir ):
  output_img_dir = os.path.join(output_dir, 'imgs/')
  output_express_dir = os.path.join(output_dir, 'expression/')
  output_pause_dir = os.path.join(output_dir, 'pausing/')

  setDir( output_dir )
  os.mkdir( output_img_dir )
  os.mkdir( output_express_dir )
  os.mkdir( output_pause_dir )
  write_tmp_variable("output_dir", output_dir)
  write_tmp_variable("output_img_dir", output_img_dir)
  write_tmp_variable("output_express_dir", output_express_dir)
  write_tmp_variable("output_pause_dir", output_pause_dir)

def remove_tmp():
  # 所有tmp开头的文件删掉
  for file in os.listdir('./'):
    if file.startswith('tmp_'):
      os.remove(file)

def add_string_to_file(string, file):
  with open(file, 'a+') as f:
    if not string.endswith('\n'): string = string + '\n'
    f.write(string )

def write_error(string):
  output = read_tmp_variable('output_dir')
  add_string_to_file(string, os.path.join( output,'./error.log'))
  print(string)
  sys.exit()

def write_log(string):
  output = read_tmp_variable('output_dir')
  add_string_to_file(string, os.path.join( output,'./process.log'))
  print(string)

def write_warning(string):
  output = read_tmp_variable('output_dir')
  add_string_to_file(string, os.path.join( output,'./warning.log'))
  print(string)

def print_error( error_str):
  print(error_str)
  sys.exit()

def filter_list_none(raw_list):
  return list(filter(None, raw_list ))

def copy_dir(source_dir, target_dir):
    source_path = os.path.abspath(source_dir)
    target_path = os.path.abspath( target_dir )
    if not os.path.exists(target_path):
      os.makedirs(target_path)

    if os.path.exists(source_path):
      for root, dirs, files in os.walk(source_path):
        for file in files:
          src_file = os.path.join(root, file)
          shutil.copy(src_file, target_path)

def make_zip(source_dir, output_filename):
    # zipf = zipfile.ZipFile(output_filename, 'w')
    # pre_len = len(os.path.dirname(source_dir))
    # for parent, dirnames, filenames in os.walk(source_dir):
    #     for filename in filenames:
    #         pathfile = os.path.join(parent, filename)
    #         arcname = pathfile[pre_len:].strip(os.path.sep)     #相对路径
    #         zipf.write(pathfile, arcname)
    # zipf.close()
    # 上面压缩，浏览器检测错误，无法下载
    os.system(' tar czvf %s %s'%(output_filename, source_dir) )

def setDir(path):
  folder = os.path.exists(path)
  if folder:
    shutil.rmtree(path)
  os.makedirs(path)

def zip_plus_move(output_dir, output_gz_file, mv_dir):
  make_zip(output_dir, output_gz_file)
  try:
    shutil.move(output_gz_file, mv_dir)
  except:
    pass

def run_bash_script(cmd, stdout_file=None, shell = True):
  '''
  Run a command
  如果返回值以Error: 开头就是运行失败, 反之就运行成功, 返回值就是output
  程序需要输出什么文本结果 根据指定的参数, 不在这个函数的控制范围内
  '''
  if not stdout_file:
    # 把 output 输出到控制台
    process = subprocess.Popen(cmd, shell = shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
  else:
    with open( stdout_file, 'w') as f:
      process = subprocess.Popen(cmd, shell = shell, stdout = f, stderr=subprocess.PIPE)

  stdout, stderr = process.communicate()
  return_code = process.poll()
  if return_code != 0:
    # 把 错误信息 打印到 控制台
    print("Process failed!")
    print(cmd)
    print(stderr)
    return('Error: ' + stderr)
  return(stdout)

def run_in_parallel(input_list, args, func, kwargs_dict = None, cpus = None, onebyone = False):
    '''
    Take an input list, divide into chunks and then apply a function to each of the chunks in parallel.
    '''
    if not cpus:
        #divide by two to get the number of physical cores
        #subtract one to leave one core free
        cpus = int(os.cpu_count()/2 - 1)  # 12超线程, 算出来cpus为5
    elif cpus == "all":
        cpus = os.cpu_count()
    # 一个无用参数的位置用于标记 第几轮并行
    arg_to_parallelize = args.index("foo")
    if not onebyone:
        # 列表中0-20，chunk_list是 [[0, 5, 10, 15], [1, 6, 11, 16], [2, 7, 12, 17], [3, 8, 13, 18], [4, 9, 14, 19]]
        chunk_list = [input_list[i::cpus] for i in range(cpus)]
    else:
        #each element in the input list will constitute a chunk of its own.
        chunk_list = input_list
    pool = multiprocessing.Pool(cpus)
    results = []
    for i in chunk_list:
        current_args = args.copy()
        # current_args = copy.deepcopy(args)
        # 把input数据放进参数中
        current_args[arg_to_parallelize] = i
        if kwargs_dict:
            process = pool.apply_async(func, tuple(current_args), kwargs_dict)
        else:
            process = pool.apply_async(func, tuple(current_args))
        results.append(process)
    pool.close()
    pool.join()
    return(results)

def is_tool(name):
  """Check whether `name` is on PATH and marked as executable."""
  # from whichcraft import which
  return which(name) is not None


import pandas as pd
from scipy import stats
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from parse_gtf import get_gene_df

def scatterplot( x_list, y_list, title, x_label, y_label, text, output_file ):
  fig = plt.figure(figsize=(12, 7))
  ax = plt.gca()
  plt.scatter(x=x_list, y=y_list)
  plt.text(0.5, 0.75, text, transform = ax.transAxes)
  plt.title( title )
  plt.xlabel(x_label)
  plt.ylabel(y_label)
  fig.savefig(output_file)
  plt.close(fig)

output_root = '../../nasap_GSM3309956/'
gtf = '../../../data/Homo_sapiens.GRCh38.93.gtf'


print( protein_data.head() )
filter_genes = protein_data[pd.to_numeric( protein_data.pi, errors='coerce' ) > 2]


gene_df = get_gene_df( gtf )
print( gene_df.head )
filter_gene_df = pd.DataFrame({})
forward_bw_signal = pyBigWig.open(  forward_bw )
reverse_bw_signal = pyBigWig.open(  reverse_bw )




import os,  subprocess, fire, psutil, time

from loguru import logger

# from .libs.py_ext import
# 思路:
# 每个命令都有参数，参数是值或文件(文件夹), 每个参数都要被检验
# 参数可以是required 或者 非required
# 如果是文件则检查文件是否存在
# 文件夹如果不存在，并创建

self_src = os.path.abspath( os.path.join(os.path.dirname(__file__)) ) + '/scripts/'
global_root = './'

def _ver():
  with open(os.path.abspath(os.path.join(os.path.dirname(__file__) ) ) + '/__version__.py') as f:
    ver = f.read().split('version=')[1].replace('"', '').strip()

  print("nASAP version is " + ver)

def _write_err( content, output_root=None):
  logger.error( content )
  if not output_root: output_root = global_root
  with open( output_root + 'error.log', 'w') as f:
    f.write( content )
  os.sys.exit(0)

def _write_info(content, output_root):
  logger.info(content)
  with open( output_root + 'info.log', 'a') as f:
    f.write("\n")
    f.write( content )

def _check_soft(name_list, isPython=False):
  if isPython:
    for name in name_list:
      try:
        exec('import '+ name )
      except:
        if name == 'community':
          name = 'python-louvain'
        _write_err( 'import %s error.\nyou need to install %s first'%(name, name) )
  else:
    for name in name_list:
      soft_dep = os.system(name + ' --version')
      if soft_dep != 0:
        _write_err( 'import %s error.\nyou need to install %s first'%(name, name) )

def _check_para(name, value, type, isFile=False, required=False, output_root='./test_output/'):
  # 先检查参数的类型，
  # 如果是文件就检查文件是否存在
  # 如果是文件夹就判断文件夹是否存在，不存在就创建
  if not output_root.endswith('/'): output_root = output_root + '/'
  if required:
    if value == None:
      _write_err(name + ' is a required parameter')
  if not isinstance(value, type):
    _write_err('Error: '+ name + ' should be a ', type, ' type.')
  if isFile:
    if not os.path.exists(value):
      _write_err('Error: ' + value + ' not exist!')
  return True

def _get_cores(value):
  total_cores = psutil.cpu_count()
  cur_cores = 1
  if isinstance(value, int):
    if value > total_cores:
      print("warning: your cpu cores number is", total_cores, "we set the parameter cores to", total_cores)
      cur_cores = total_cores
    else:
      cur_cores = value
    return cur_cores
  if isinstance(value, str):
    if str.lower() == 'max':
      cur_cores = total_cores
      return cur_cores
    if str.lower == 'max/2':
      cur_cores = total_cores/2
      return cur_cores
  print("warning: cores parameter error use default cores=1")
  return 1

def _cur_time():
  return time.asctime( time.localtime(time.time()) )

def _create_project_dir(output_root):
  global global_root
  global_root = output_root
  if not output_root.endswith('/'): output_root = output_root + '/'
  json_dir=output_root + 'json/'
  fq_dir=output_root + 'fastq/'
  txt_dir=output_root + 'txt/'
  img_dir=output_root + 'imgs/'
  sam_dir=output_root + 'sam/'
  bed_dir=output_root + 'bed/'
  bw_dir=output_root + 'bw/'
  csv_dir=output_root + 'csv/'
  html_dir=output_root + 'html/'
  for d in [output_root, json_dir, fq_dir, txt_dir, img_dir, sam_dir, bed_dir, bw_dir, csv_dir, html_dir]:
    if not os.path.exists(d):
      os.makedirs(d)

def _check_output(output_root):
  output_subdir = ['json/','fastq/','txt/', 'imgs/','sam/','bed/','bw/', 'html/']

  for sub_dir in output_subdir:
    if not output_root.endswith('/'): output_root = output_root + '/'
    sub_dir = output_root + sub_dir
    if not os.listdir(sub_dir):
      os.removedirs(sub_dir)


def server(output_root='./test_output/', forward_bw=None, reverse_bw=None, gtf=None,  cores=1, tf_source=None, enhancer_source=None, select_nodes_file=None, express_file=None):
  global global_root
  global_root = output_root
  # logger.info('server--construct output dir')
  if not output_root.endswith('/'): output_root = output_root + '/'
  _create_project_dir(output_root)
  logger.info('server--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'numpy', 'scipy', 'networkx','community', 'hvplot'], isPython=True)



  logger.info('server--check parameter and input files')

  feature_assign_cmd_list = ['python', self_src + 'feature_attrs.py']
  pausing_sites_cmd_list = ['python', self_src + 'pausing_sites_low_memory.py']
  network_cmd_list = ['python', self_src + 'network_analysis.py']
  render_cmd_list = ['python', self_src + 'render_template.py']
  render_cmd_list.extend(['--type', 'server'])

  feature_assign_cmd_list.extend(['--output_root', output_root])
  pausing_sites_cmd_list.extend(['--output_root', output_root])
  network_cmd_list.extend(['--output_root', output_root])
  render_cmd_list.extend(['--output_root', output_root])


  # 检查必须参数
  # 检查文件
  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--forward_bw', forward_bw] )
  pausing_sites_cmd_list.extend(['--forward_bw', forward_bw] )
  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--reverse_bw', reverse_bw] )
  pausing_sites_cmd_list.extend(['--reverse_bw', reverse_bw] )
  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  feature_assign_cmd_list.extend(['--gtf', gtf] )

  # 可选参数
  cur_cores = _get_cores(cores)
  pausing_sites_cmd_list.extend(['--cores', str(cur_cores)])

  if not (tf_source or enhancer_source):
    logger.error( 'Error, you have to provide at least one parameter for tf_source or enhancer_source')
    os.sys.exit(1)

  if tf_source:
    _check_para('--tf_source', tf_source, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--tf_source', tf_source])

  if enhancer_source:
    _check_para('--enhancer_source', enhancer_source, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--enhancer_source', enhancer_source])
  if select_nodes_file:
    _check_para('--select_nodes_file', select_nodes_file, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--select_nodes_file', select_nodes_file])
  if express_file:
    _check_para('--express_file', express_file, str, isFile=True, required=True, output_root=output_root)
    network_cmd_list.extend( ['--express_file', express_file])

  all_steps = [feature_assign_cmd_list, pausing_sites_cmd_list, network_cmd_list, render_cmd_list]
  steps_name = ['feature_assign', 'pausing_sites', 'network_analysis', 'render_output']
  for cmd_list, step in zip(all_steps, steps_name):
    try:
      logger.info(step + ' start')
      os.system( ' '.join(cmd_list) )
      logger.info(step + ' finished')
    except Exception as e:
      _write_err(step + ' --Failed\n' + str(e) )

def assessment(output_root='./test_output/', read1=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi_loc=None, umi_len=None, bowtie_index=None, gtf=None, genome=None, scale_factor=None):
  global global_root
  global_root = output_root
  # logger.info('assessment--construct output dir')
  if not output_root.endswith('/'): output_root = output_root + '/'
  _create_project_dir(output_root)
  logger.info('assessment--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python', 'bowtie2', 'samtools', 'bedtools', 'deeptools'], isPython=False)



  # 这里不检查 文件是否存在，因为在_pp函数中有检查
  _pp(output_root=output_root, read1=read1, read2=read2, adapter1=adapter1, adapter2=adapter2, umi_loc=umi_loc, umi_len=umi_len, genome=genome)
  if read2:
    _align(output_root=output_root, read1=output_root + 'fastq/clean_read1.fq.gz', read2=output_root + 'fastq/clean_read2.fq.gz', bowtie_index=bowtie_index, gtf=gtf, cores=cores )
  else:
    _align(output_root=output_root, read1=output_root + 'fastq/clean_read1.fq.gz', bowtie_index=bowtie_index, gtf=gtf, cores=cores )

  # remove umi with gencore
  if umi_loc:
    # 不用check 值, 因为上面check 过了
    try:
      subprocess.run(self_src + 'gencore/gencore', shell=True, check=True)
      gencore_dir = self_src + 'gencore/gencore'
    except:
      _check_soft( ['gencore'], isPython=False)
      gencore_dir = 'gencore'
    try:
      os.system(' '.join([gencore_dir, '-i', output_root + 'sam/uniquemapped_sort.bam', '-o',
      output_root + 'sam/uniquemapped_sort_umi.bam', '-r', genome]) )
      # 排序 index
      os.system('samtools sort -@ %s -o %s %s'%(cores, output_root + 'sam/uniquemapped_sort.bam', output_root + 'sam/uniquemapped_sort_umi.bam') )
      os.system('samtools index -@ %s %s'%(cores, output_root + 'sam/uniquemapped_sort.bam') )
    except Exception as e:
      _write_err('gencore remove umi--Failed\n' + str(e) )

  if scale_factor:
    try:
      scale_factor = float(scale_factor)
    except Exception as e:
      _write_err('--scale_factor parsed--Failed, scale-factor must be int or float number\n' + str(e) )

  _tracks(output_root=output_root, bam=output_root + 'sam/uniquemapped_sort.bam', cores=cores, scale_factor=scale_factor)
  _render_template(output_root=output_root, type='assessment', is_server='No')

def all(output_root='./test_output/', read1=None, bowtie_index=None,  gtf=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi_loc=None, umi_len=None, tf_source=None, select_nodes_file=None, enhancer_source=None, express_file=None):
  global global_root
  global_root = output_root
  logger.info('all--check installed software')
  if not output_root.endswith('/'): output_root = output_root + '/'
  _create_project_dir(output_root)
  _check_soft( ['fastp', 'bioawk', 'python', 'bowtie2', 'samtools', 'bedtools', 'deeptools'], isPython=False)
  _check_soft( ['pandas', 'pyBigWig', 'numpy', 'scipy', 'networkx','community', 'hvplot'], isPython=True)
  assessment(output_root=output_root, read1=read1,  cores=cores, read2=read2, adapter1=adapter1, adapter2=adapter2, umi_loc=umi_loc, umi_len=umi_len, bowtie_index=bowtie_index, gtf=gtf)
  server(output_root=output_root, forward_bw=output_root + 'bw/forward.bw', reverse_bw=output_root + 'bw/reverse.bw', gtf=gtf, cores=cores, tf_source=tf_source, select_nodes_file=select_nodes_file, enhancer_source=enhancer_source, express_file=express_file )
  _render_template(output_root=output_root, type='all', is_server='No')


def _pp(output_root=os.getcwd()+'/tmp_output', read1=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi_loc=None, umi_len=None, genome=None, scaleFactor=None):
  logger.info('preprocess--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python'], isPython=False)
  if read2:
    _check_soft( ['flash'], isPython=False)

  logger.info('preprocess--check parameter and input files')
  if not output_root.endswith('/'):
    output_root = output_root + '/'
  logger.info('preprocess--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'w')
  cmd_list = ['bash', self_src + 'preprocess.bash']
  cmd_list.extend(['--output_root', output_root])
  # 必须参数 参数什么都不加
  _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--read1', read1])
  log_file.write( 'read1--'+os.path.basename(read1)+'\n' )
  log_file.write('read1_size--' + str(os.stat(read1).st_size / (1024 * 1024)) + 'Mb'+'\n')

  # 可选参数
  cur_cores = _get_cores(cores)
  cmd_list.extend(['--cores', str(cur_cores)])

  if read2:
    _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--read2', read2])
    log_file.write( 'read2--'+os.path.basename(read2)+'\n' )
    log_file.write('read2_size--' + str(os.stat(read2).st_size / (1024 * 1024)) + 'Mb'+'\n')

  if umi_loc:
    if genome:
      _check_para('--genome', genome, str, isFile=True, required=True, output_root=output_root)
    else:
      logger.error('umi module must provide genome file with --genome')
      log_file.write( 'umi module must provide genome file with --genome' )
      os.sys.exit(1)

    # 不用check_para, 因为会检测 值必须是在 列表内的值
    if umi_loc not in ['read1', 'read2', 'per_read', 'index1', 'index2', 'per_index']:
      logger.error('umi_loc can be one of index1/index2/read1/read2/per_index/per_read')
      log_file.write( 'umi_loc can be index1/index2/read1/read2/per_index/per_read'+os.path.basename(read2)+'\n' )
      os.sys.exit(1)
    else:
      if umi_loc in ['read1', 'read2', 'per_read']:
        if isinstance(umi_len, int):
          # read 模式
          cmd_list.extend(['--umi_loc', umi_loc])
          cmd_list.extend(['--umi_len', str(umi_len)])
        else:
          logger.error('umi on ' + umi_loc + ' must provide --umi_len')
          os.sys.exit(1)
      else:
        # index模式
        cmd_list.extend(['--umi_loc', umi_loc])

  if adapter1:
    _check_para('--adapter1', adapter1, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter1', adapter1])
    log_file.write( 'adapter1--'+ adapter1 +'\n' )

  if adapter2:
    _check_para('--adapter2', adapter2, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter2', adapter2])
    log_file.write( 'adapter2--'+ adapter2 +'\n' )

  # print( ' '.join(cmd_list) )
  log_file.write(_cur_time()+ 'preprocess start'+'\n')
  logger.info('preprocess--running script')
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('preprocess--Failed\n' + str(e) )

  # extract preprocess
  try:
    os.system( ' '.join(['python', self_src + 'extract_preprocess.py', output_root]) )
  except Exception as e:
    _write_err('extract preprocess--Failed\n' + str(e) )

  log_file.close()
  logger.success(_cur_time() + 'preprocess--Finished. Find the results in ' + output_root)

def preprocess_fast(output_root=os.getcwd()+'/tmp_output', read1=None,  cores=1, read2=None, adapter1=None, adapter2=None, umi=None, bowtie_index=None):
  global global_root
  global_root = output_root
  logger.info('preprocess--check installed software')
  _check_soft( ['fastp', 'bioawk', 'python'], isPython=False)
  if read2:
    _check_soft( ['flash'], isPython=False)

  logger.info('preprocess--check parameter and input files')
  if not output_root.endswith('/'):
    output_root = output_root + '/'
  logger.info('preprocess--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'w')
  cmd_list = ['bash', self_src + 'preprocess_fast.bash']
  cmd_list.extend(['--output_root', output_root])
  # 必须参数 参数什么都不加
  _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--read1', read1])
  _check_para('--bowtie_index', bowtie_index, str, required=True, output_root=output_root)
  cmd_list.extend(['--bowtie_index', bowtie_index])

  # 可选参数
  # cur_cores = _get_cores(cores)
  cur_cores = cores
  cmd_list.extend(['--cores', str(cur_cores)])

  if read2:
    _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--read2', read2])

  if adapter1:
    _check_para('--adapter1', adapter1, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter1', adapter1])

  if adapter2:
    _check_para('--adapter2', adapter2, str, required=True, output_root=output_root)
    cmd_list.extend(['--adapter2', adapter2])

  # print( ' '.join(cmd_list) )
  log_file.write(_cur_time()+ 'preprocess start'+'\n')
  logger.info('preprocess--running script')
  print( ' '.join(cmd_list) )
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('fast preprocess--Failed\n' + str(e) )

  # todo 没有uniquemapped_sort.bam
  _tracks(output_root=output_root, bam=output_root + 'sam/uniquemapped_sort.bam')

  log_file.close()
  logger.success(_cur_time() + 'preprocess--Finished. Find the results in ' + output_root)


def _align(read1=None, bowtie_index=None, gtf=None, output_root=None, cores=1, read2=None ):
  logger.info('alignment--check installed software')
  _check_soft( ['bowtie2', 'samtools', 'bedtools'], isPython=False)

  logger.info('alignment--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('alignment--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'a+' )
  cmd_list = ['bash', self_src + 'map1.bash']
  cmd_list.extend(['--output_root', output_root])

  _check_para('--read1', read1, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--read1', read1])
  log_file.write('clean_read1--'+os.path.basename(read1) + '\n' )
  log_file.write('clean_read1_size--' + str(os.stat(read1).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--bowtie_index', bowtie_index, str, required=True, output_root=output_root)
  cmd_list.extend(['--bowtie_index', bowtie_index])

  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--gtf', gtf])


  cur_cores = _get_cores(cores)
  cmd_list.extend(['--cores', str(cur_cores)])
  if read2:
    _check_para('--read2', read2, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--read2', read2])
    log_file.write('clean_read2--'+os.path.basename(read2) + '\n' )
    log_file.write('clean_read2_size--' + str(os.stat(read2).st_size / (1024 * 1024)) + 'Mb\n')

  log_file.write(_cur_time()+ 'alignment start\n')
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('original_mapping--Failed\n' + str(e) )

  try:
    os.system( ' '.join(['python', self_src + 'map_split.py',
    '--sam_dir='+output_root+'sam/', '--sam_file=original.sam',
    '--output_root='+output_root
    ]) )
  except Exception as e:
    _write_err('map_split--Failed\n' + str(e) )

  cmd_list[1] = self_src + 'map2.bash'
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('map_comutation--Failed\n' + str(e) )

  log_file.close()
  logger.success(_cur_time()+ 'alignment--Finished. Find the results in ' + output_root)

def _bamCov(bam=None, output_root=None, scale_factor=None, cores=None, five_end=None):
  logger.info('genome_tracks--check installed software')
  _check_soft( ['deeptools'], isPython=False)

  logger.info('genome_tracks--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('genome_tracks--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'a+' )

  cmd_list = ['bamCoverage', '--binSize', '1']
  _check_para('--bam', bam, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--bam', bam])
  log_file.write('bam--'+os.path.basename(bam) + '\n' )
  bam_size = os.stat(bam).st_size / (1024 * 1024)
  log_file.write('bam_size--' + str(bam_size) + 'Mb\n')
  if cores:
    cur_cores = _get_cores(cores)
    cmd_list.extend(['-p', str(cur_cores)])
  else:
    # 这里运行时间太长 一开始cores直接写死 40，40是线程 后来发现小文件不合适
    if bam_size < 100:
      cmd_list.extend(['-p', str(4)])
    elif bam_size < 500:
      cmd_list.extend(['-p', str(8)])
    elif bam_size < 1000:
      cmd_list.extend(['-p', str(16)])
    else:
      cmd_list.extend(['-p', str(32)])

  if scale_factor:
    cmd_list.extend(['--scaleFactor', str(scale_factor)])

  cmd_list_forward, cmd_list_reverse = cmd_list.copy(), cmd_list.copy()
  cmd_list_forward.extend(['--filterRNAstrand', 'forward'])
  cmd_list_reverse.extend(['--filterRNAstrand', 'reverse'])
  cmd_list_forward.extend(['-o', output_root + 'bw/forward.bw'])
  cmd_list_reverse.extend(['-o', output_root + 'bw/reverse.bw'])
  log_file.write(_cur_time()+ ': generate forward genome track start\n')
  try:
    subprocess.run(' '.join(cmd_list_forward),  shell=True,  check=True)
    if five_end:
      cmd_list_forward[-1] = output_root + 'bw/forward_5_end.bw'
      cmd_list_forward.extend(['--Offset', '1'])
      subprocess.run(' '.join(cmd_list_forward),  shell=True,  check=True)
  except Exception as e:
    _write_err('forward bamCoverage--Failed\n' + str(e) )

  log_file.write(_cur_time()+ ': generate reverse genome track start\n')
  try:
    subprocess.run(' '.join(cmd_list_reverse),  shell=True,  check=True)
    if five_end:
      cmd_list_reverse[-1] = output_root + 'bw/reverse_5_end.bw'
      cmd_list_reverse.extend(['--Offset', '1'])
      subprocess.run(' '.join(cmd_list_reverse),  shell=True,  check=True)
  except Exception as e:
    _write_err('reverse bamCoverage--Failed\n' + str(e) )

  log_file.close()
  logger.success(_cur_time()+'genome_tracks--Finished. Find the results in ' + output_root)


def _genomeCov(bam=None, output_root=None, scale_factor=None, cores=None, five_end=None):
  # BAM -> BedGraph -> BigWig
  # 1 software test
  logger.info('genome_tracks--check installed software')
  _check_soft(['bedtools'], isPython=False)
  _check_soft(['samtools'], isPython=False)
  try:
    if os.system(self_src + 'ucsc/bedGraphToBigWig') != 0:
      bg2bw_dir = self_src + 'ucsc/bedGraphToBigWig'
    else:
      if os.system('bedGraphToBigWig') != 0:
        bg2bw_dir = 'bedGraphToBigWig'
      else:
        _write_err( 'import %s error.\nyou need to install %s first'%('bedGraphToBigWig', 'bedGraphToBigWig') )
  except:
    if os.system('bedGraphToBigWig') != 0:
      bg2bw_dir = 'bedGraphToBigWig'
    else:
      _write_err( 'import %s error.\nyou need to install %s first'%('bedGraphToBigWig', 'bedGraphToBigWig') )

  # 2 bedtools genomecov -ibam filename.bam -bg > filename.bedgraph
  logger.info('genome_tracks--check parameter and input files')
  if not output_root.endswith('/'): output_root = output_root + '/'
  logger.info('genome_tracks--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'a+' )

  cmd_list = ['bedtools', 'genomecov', '-bg']
  _check_para('--bam', bam, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['-ibam', bam])
  log_file.write('bam--'+os.path.basename(bam) + '\n' )
  bam_size = os.stat(bam).st_size / (1024 * 1024)
  log_file.write('bam_size--' + str(bam_size) + 'Mb\n')
  def generate_chr_size(bam):
    os.system('samtools idxstats %s > %stmp_chr_size.txt'%(bam, output_root) )
    with open( output_root + 'chr_size.txt', 'w') as f:
      for ln in open(output_root + 'tmp_chr_size.txt'):
        if '*' in ln: continue
        seq_id, seq_len, _, _ = ln.split('\t')
        f.write(seq_id + '\t' + seq_len + '\n')
  generate_chr_size(bam)
  if scale_factor:
    cmd_list.extend(['-scale', str(scale_factor)])

  cmd_list_forward, cmd_list_reverse = cmd_list.copy(), cmd_list.copy()
  cmd_list_forward.extend(['-strand', '+'])
  cmd_list_reverse.extend(['-strand', '-'])
  forward_bg_file = output_root + 'tmp_forward.bedgraph'
  reverse_bg_file = output_root + 'tmp_reverse.bedgraph'
  cmd_list_forward.extend(['| sort -k1,1 -k2,2n', '>', forward_bg_file])
  cmd_list_reverse.extend(['| sort -k1,1 -k2,2n', '>', reverse_bg_file])
  log_file.write(_cur_time()+ ': generate forward genome track start\n')
  # 3 bedGraphToBigWig filename.bedgraph $dm6 filename.bw
  def bg2bw(bg_file, bw_file):
    bg2bw_cmd_list =[bg2bw_dir, bg_file, output_root + 'chr_size.txt', output_root + 'bw/' + bw_file]
    subprocess.run(' '.join(bg2bw_cmd_list),  shell=True,  check=True)

  try:
    subprocess.run(' '.join(cmd_list_forward),  shell=True,  check=True)
    bg2bw(forward_bg_file, 'forward.bw')
    if five_end:
      forward_bg_file = output_root + 'tmp_forward_5_end.bedgraph'
      cmd_list_forward[-1] = output_root + forward_bg_file
      cmd_list_forward.extend(['-5'])
      subprocess.run(' '.join(cmd_list_forward),  shell=True,  check=True)
      bg2bw(forward_bg_file, 'forward_5_end.bw')
  except Exception as e:
    _write_err('forward bamCoverage--Failed\n' + str(e) )

  log_file.write(_cur_time()+ ': generate reverse genome track start\n')
  try:
    subprocess.run(' '.join(cmd_list_reverse),  shell=True,  check=True)
    bg2bw(reverse_bg_file, 'reverse.bw')
    if five_end:
      reverse_bg_file = output_root + 'tmp_reverse_5_end.bedgraph'
      cmd_list_reverse[-1] = output_root + reverse_bg_file
      cmd_list_reverse.extend(['-5'])
      subprocess.run(' '.join(cmd_list_reverse),  shell=True,  check=True)
      bg2bw(reverse_bg_file, 'reverse_5_end.bw')
  except Exception as e:
    _write_err('reverse bamCoverage--Failed\n' + str(e) )

  log_file.close()
  logger.success(_cur_time()+'genome_tracks--Finished. Find the results in ' + output_root)


def _tracks(bam=None, output_root=None, cores=None, scale_factor=None, five_end=None):
  bam_size = float(os.path.getsize(bam)/1000000)
  if bam_size < 50:
    _genomeCov(bam=bam, output_root=output_root, scale_factor=scale_factor,  five_end=five_end)
  else:
    _bamCov(bam=bam, output_root=output_root, scale_factor=scale_factor, cores=cores,  five_end=five_end)


def feature_assign(forward_bw=None, reverse_bw=None, gtf=None, output_root=None, tss_up=150, tss_down=150, gene_length=2000, rpkm_threshold=0.1):
  global global_root
  global_root = output_root
  if not output_root.endswith('/'): output_root = output_root + '/'
  # logger.info('feature_assign--construct output dir')
  _create_project_dir(output_root)
  logger.info('feature_assign--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'scipy'], isPython=True)

  logger.info('feature_assign--check parameter and input files')

  log_file = open(output_root + 'tmp.log', 'w' )

  cmd_list = ['python', self_src + 'feature_attrs.py']
  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--forward_bw', forward_bw])
  log_file.write('forward_bw--'+os.path.basename(forward_bw) + '\n' )
  log_file.write('forward_bw_size--' + str(os.stat(forward_bw).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--reverse_bw', reverse_bw])
  log_file.write('reverse_bw--'+os.path.basename(reverse_bw) + '\n' )
  log_file.write('reverse_bw_size--' + str(os.stat(reverse_bw).st_size / (1024 * 1024)) + 'Mb\n')

  _check_para('--gtf', gtf, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--gtf', gtf])

  cmd_list.extend(['--output_root', output_root])
  log_file.write(_cur_time()+ 'feature assign start\n')


  try:
    tss_up = int(tss_up)
    cmd_list.extend(['--tss_up', str(tss_up)])
  except Exception as e:
    _write_err('--tss_up parsed--Failed, tss_up must be int number\n' + str(e) )
  try:
    tss_down = int(tss_down)
    cmd_list.extend(['--tss_down', str(tss_down)])
  except Exception as e:
    _write_err('--tss_down parsed--Failed, tss_down must be int number\n' + str(e) )

  try:
    gene_length = int(gene_length)
    cmd_list.extend(['--gene_length', str(gene_length)])
  except Exception as e:
    _write_err('--gene_length parsed--Failed, gene_length must be int number\n' + str(e) )
  try:
    rpkm_threshold = float(rpkm_threshold)
    cmd_list.extend(['--rpkm_threshold', str(rpkm_threshold)])
  except Exception as e:
    _write_err('--rpkm_threshold parsed--Failed, rpkm_threshold must be int number\n' + str(e) )

  try:
    os.system( ' '.join(cmd_list) )
    logger.success('feature_assign--Finished. Find the results in ' + output_root)
  except Exception as e:
    _write_err('feature_assign--Failed\n' + str(e) )

  log_file.write(_cur_time()+ 'feature assign finished\n')
  log_file.close()
  _render_template(output_root=output_root, type='quantification', is_server='No')

def pausing_sites(forward_bw=None, reverse_bw=None, output_root=None, cores=1):
  global global_root
  global_root = output_root
  if not output_root.endswith('/'): output_root = output_root + '/'
  # logger.info('pausing_sites--construct output dir')
  _create_project_dir(output_root)
  logger.info('pausing_sites--check installed software')
  _check_soft( ['pandas', 'pyBigWig', 'numpy'], isPython=True)

  logger.info('pausing_sites--check parameter and input files')

  log_file = open(output_root + 'tmp.log', 'w' )

  cmd_list = ['python', self_src + 'pausing_sites_low_memory.py']

  _check_para('--forward_bw', forward_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--forward_bw', forward_bw])
  log_file.write('forward_bw--'+os.path.basename(forward_bw) )
  log_file.write('forward_bw_size--' + str(os.stat(forward_bw).st_size / (1024 * 1024)) + 'Mb')

  _check_para('--reverse_bw', reverse_bw, str, isFile=True, required=True, output_root=output_root)
  cmd_list.extend(['--reverse_bw', reverse_bw])
  log_file.write('reverse_bw--'+os.path.basename(reverse_bw) )
  log_file.write('reverse_bw_size--' + str(os.stat(reverse_bw).st_size / (1024 * 1024)) + 'Mb')

  cmd_list.extend(['--output_root', output_root])
  if cores:
    cur_cores = _get_cores(cores)
    cmd_list.extend(['--cores', str(cur_cores)])

  log_file.close()
  try:
    os.system( ' '.join(cmd_list) )
    logger.success('pausing_sites--Finished. Find the results in ' + output_root)
  except Exception as e:
    _write_err('pausing_sites--Failed\n' + str(e) )

  _render_template(output_root=output_root, type='pausing', is_server='No')

def network_analysis(tf_source=None, enhancer_source=None, select_nodes_file=None, express_file=None, output_root=None ):
  global global_root
  global_root = output_root
  if not output_root.endswith('/'): output_root = output_root + '/'
  _create_project_dir(output_root)
  # logger.info('network_analysis--construct output dir')
  logger.info('network_analysis--check installed software')
  _check_soft(['networkx','community', 'hvplot', 'numpy'], isPython=True)

  cmd_list = ['python', self_src + 'network_analysis.py']
  logger.info('network_analysis--check parameter and input files')
  cmd_list.extend(['--output_root', output_root])

  log_file = open(output_root + 'tmp.log', 'w' )

  if not(tf_source or enhancer_source):
    _write_err( 'network_analysis--Failed ' + 'no tf or enhancer source file')
    os.sys.exit(1)

  if tf_source:
    _check_para('--tf_source', tf_source, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--tf_source', tf_source])
  if select_nodes_file:
    _check_para('--select_nodes_file', select_nodes_file, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--select_nodes_file', select_nodes_file])
  if enhancer_source:
    _check_para('--enhancer_source', enhancer_source, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--enhancer_source', enhancer_source])
  if express_file:
    _check_para('--express_file', express_file, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--express_file', express_file])

  try:
    os.system( ' '.join(cmd_list) )
    logger.success('network analysis--Finished. Find the results in ' + output_root)
    _render_template(output_root=output_root, type='network', is_server='No')
  except Exception as e:
    _write_err('network analysis--Failed\n' + str(e) )

def network_links(specie=None, region='', gtf=None, forward_bw=None, reverse_bw=None, rpkm_file=None, output_root=None):
  global global_root
  global_root = output_root
  logger.info('network_links--check installed software')
  _check_soft(['pyGenomeTracks'], isPython=False)
  logger.info('network_links--check parameter and input files')
  if not output_root.endswith('/'):
    output_root = output_root + '/'
  logger.info('network_links--construct output dir')
  _create_project_dir(output_root)
  log_file = open(output_root + 'tmp.log', 'w')
  cmd_list = ['python', self_src + 'genome_track_visualization.py']
  cmd_list.extend(['--output_root', output_root])
  # 必须参数 参数什么都不加
  if not specie:
    print( 'you must provide specie')
    os.sys.exit(1)
  if region == '':
    print( 'you must provide regulatory region with --region i.e --region chr1:1:5000000')
    os.sys.exit(1)
  if not gtf:
    print( 'you must provide gtf file')
    os.sys.exit(1)
  if not forward_bw:
    print( 'you must provide forward bw file')
    os.sys.exit(1)
  if not reverse_bw:
    print( 'you must provide reverse bw file')
    os.sys.exit(1)

  cmd_list.extend(['--specie', specie])
  cmd_list.extend(['--region', region])
  cmd_list.extend(['--gtf', os.path.abspath(gtf)])
  cmd_list.extend(['--forward', os.path.abspath(forward_bw)])
  cmd_list.extend(['--reverse', os.path.abspath(reverse_bw)])
  # 非必须参数
  if rpkm_file:
    _check_para('--rpkm_file', rpkm_file, str, isFile=True, required=True, output_root=output_root)
    cmd_list.extend(['--rpkm_file', rpkm_file])
  print( ' '.join(cmd_list) )
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('network links--Failed\n' + str(e) )

def _render_template(type='server', output_root=None, is_server=None):
  global global_root
  global_root = output_root
  # 每个步骤对应了输出文件， 用来丰富报表的数据
  # 考虑到 要和论文的内容对应上，所以不能按照 流程来生成报告
  # 把报告分成 1 基础信息 2 qc 3 转录级别评价 4 暂停因子和暂停位点 5 网络分析
  # 只有两种情况 会用到render_template
  logger.info('render_template--check installed software')
  cmd_list = ['python', self_src + 'render_template.py']

  cmd_list.extend(['--type', type])
  if not output_root.endswith('/'): output_root = output_root + '/'
  _create_project_dir(output_root)
  cmd_list.extend(['--output_root', output_root])
  if is_server:
    cmd_list.extend( ['--is_server', is_server])
  print( ' '.join(cmd_list) )
  try:
    os.system( ' '.join(cmd_list) )
  except Exception as e:
    _write_err('render template--Failed\n' + str(e) )


  # 基础信息 table
  # 输入文件名
  # Forward_bw, Reverse_bw,

def main():
  fire.core.Display = lambda lines, out: print(*lines, file=out)
  fire.Fire({
    #'preprocess': preprocess,
    #'alignment': alignment,
    #'genome_tracks': genome_tracks,
    '--version': _ver,
    '-V': _ver,
    'assessment': assessment,
    'feature_assign': feature_assign,
    'pausing_sites': pausing_sites,
    'network_analysis': network_analysis,
    'network_links': network_links,
    'render_template': _render_template,
    'server': server,
    'all': all,
    'preprocess_fast': preprocess_fast,
  })

if __name__ == '__main__':
  main()


import re, math, pickle
from collections import Counter, defaultdict

import numpy as np
import pandas as pd
from scipy import stats

import matplotlib.pyplot as plt
plt.style.use('seaborn-whitegrid')
from matplotlib.pyplot import Polygon
from matplotlib import ticker
import seaborn as sns
sns.set(style='whitegrid')

output_root = '/home/nasap_performance/GSM1480325/'
output_root = './'

'''
insertion_size_list = [int( ln.split('\t')[1].strip() ) for ln in open(output_root + 'txt/adapter_read_len1.txt')]
# 计算 y->insertion size的频率(Counter就行)
insert_size_counter = Counter(insertion_size_list)
x_list = list( insert_size_counter.keys() )
y_list = list( insert_size_counter.values() )
x_max = max(x_list)  # x_max对应reads 没有去掉 adapter
adapter_length, length_count = [], []
degraded_reads, intact_reads = 0, 0
for x, y in zip(x_list, y_list):
  if x==x_max:
    continue
  l = x_max - x
  adapter_length.append( l )
  length_count.append(y)
  if x >= 10 and x < 20:
    degraded_reads+=y
  if x >= 30 and x < 40:
    intact_reads+=y

degradation_ratio = degraded_reads/float(intact_reads)
stats_list = [] # delete later
stats_list.append( ['Degradation ratio', degradation_ratio] )
max_length =max(length_count)

data = {
  'adapter_length': adapter_length,
  'length_count': length_count,
  'max_length': max_length,
  'x_max': x_max,
  'degradation_ratio': degradation_ratio,
}

with open(output_root + 'adapter_dist.pickle', 'wb') as f:
  pickle.dump(data, f)
'''


# x->insertion size,
# y->number of reads,
# text->degradation ratio
# i scatter
# i 两段阴影
# todo draw_adapter_distribution->insertion distribution
def insertion_distribution(data, output_root):
  max_x, max_y = max(data['adapter_length']), max(data['length_count'])
  fig, ax = plt.subplots(figsize=(8, 8))
  # 散点
  ax.scatter( data['adapter_length'], data['length_count'], color='black', alpha=0.6 )
  # 阴影 用polygon
  poly1=Polygon([(0, 0),(0, max_y), (20, max_y), (20, 0)],alpha=0.3, facecolor='#FFE6D0', edgecolor="black", linestyle="-." )
  poly2=Polygon([(20, 0), (20, max_y), (30, max_y), (30, 0)],alpha=0.3, facecolor='#FFFDE5', edgecolor="black", linestyle="-."  )
  ax.add_patch(poly1)
  ax.add_patch(poly2)
  # text
  plt.text(max_x *0.5, 1, 'degradation rate:' + str(round(data['degradation_ratio'], 2) ), zorder=99 )
  plt.text(10, max_y *0.1, 'high degradation', ha="left", va="bottom", rotation=90, zorder=99)
  plt.text(25, max_y *0.1, 'partial degradation', ha="left", va="bottom", rotation=90, zorder=99)
  # ax.text(0.7, 0.9, 'degradation ratio ' + str(round(degradation_ratio, 2) ) , ha='center', va='center', transform=ax.transAxes, zorder=99, backgroundcolor='white')
  plt.xlim([0, max_x*1.05])
  plt.xlabel('Size of insertion')
  plt.ylabel('Number of reads')
  # ax.set_yscale('log')
  # plt.tight_layout()
  formatter = ticker.ScalarFormatter(useMathText=True)
  formatter.set_scientific(True)
  formatter.set_powerlimits((-1,1))
  ax.yaxis.set_major_formatter(formatter)

  plt.savefig(output_root + 'imgs/adapter_insertion_distribution.png')
  plt.savefig(output_root + 'imgs/adapter_insertion_distribution.pdf' )
  plt.close(fig)

#{'adapter_length': adapter_length,
# 'length_count': length_count,
# 'degradation_ratio': degradation_ratio}
fr = open(output_root + 'adapter_dist.pickle', 'rb')
adapter_dist_data = pickle.load(fr)
insertion_distribution( adapter_dist_data, output_root  )


# sharex -> reads_length
# y -> reads_count
# axs -> steps
# annotate -> qc
def readsLength_distribution(reads_length_data, output_root ): # table_data, table_col, table_row
  fig, axes = plt.subplots(len(reads_length_data['steps']), 1, sharex=True, sharey=True, figsize=(14, 20))

  for i, step in enumerate(reads_length_data['steps']):
    x, y = reads_length_data['bar_list'][i]
    axes[i].bar(x, y)
    axes[i].set_ylabel('' )
    axes[i].grid( axis='y')
    axes[i].set_title('Preprocess: ' + step, fontsize = 14, fontweight ='bold', verticalalignment='bottom', loc ='left')
    axes[i].text(0.01, 0.9, reads_length_data['annotate_list'][i],
        horizontalalignment='left',
        verticalalignment='top',
        transform=axes[i].transAxes)
    axes[i].spines['top'].set_visible(False)
    axes[i].spines['right'].set_visible(False)
    axes[i].spines['bottom'].set_visible(False)
    # axes[i].spines['left'].set_visible(False)
  plt.xlabel('Reads length')
  plt.yscale("log")
  sns.despine(left=True)
  plt.savefig(output_root + 'imgs/reads_distribution.png')
  plt.savefig(output_root + 'imgs/reads_distribution.pdf')
  plt.close(fig)

# {'steps': steps,
# 'bar_list': bar_list,
# 'annotate_list': [annotate_text1, annotate_text2, annotate_text3, annotate_text4, annotate_text5]}
fr = open(output_root + 'readsLength.pickle', 'rb')
reads_length_data = pickle.load(fr)
readsLength_distribution(reads_length_data, output_root)

# x-> steps
# y->q20, q30
def q20_q30_line( data, output_root):
  fig, ax = plt.subplots(figsize=(10, 4))
  x = [1, 2, 3, 4, 5]
  l1, l2 = ax.plot(x, data['q20_list'], x, data['q30_list'] , lw=2)
  l1.set_label('Q20')
  l2.set_label('Q30')
  ax.legend()
  ax.set_xticks([1, 2, 3, 4, 5])
  ax.set_xticklabels(data['steps'], fontsize=18) #tex code
  plt.xticks(rotation=30)

  plt.tight_layout()
  plt.savefig(output_root + 'imgs/quality_trend.png')
  plt.savefig(output_root + 'imgs/quality_trend.pdf')
  plt.close(fig)

q20_list = [0.975902, 0.975577, 0.993754, 0.993774, 0.99425, ]
q30_list = [0.958546, 0.959483, 0.978441, 0.9785, 0.97924 ]
q20_q30_data = {
  'q20_list': q20_list,
  'q30_list': q30_list,
  'steps': ['raw reads', 'adapter removed', 'trim', 'polyX removed', 'quality filtered']
}
# {'q20_list': q20_list,
# 'q30_list': q30_list,
# 'steps': steps}
q20_q30_line( q20_q30_data, output_root)


# x->steps
# y->x_have, x_not, x_failed
def preprocess_summary_stack(data, output_root ):
  x_have, x_not, x_failed = data['stackbar_list']
  fig, ax = plt.subplots(figsize=(8,8))
  plt.bar(data['steps'], x_have, label="reads with feature",edgecolor = 'black')
  plt.bar(data['steps'], x_not, label="reads without feature",edgecolor = 'black', bottom = x_have)
  plt.bar(data['steps'], x_failed, label="failed reads",edgecolor = 'black', bottom = [i+j for i, j in zip(x_have,x_not)])
  plt.legend( bbox_to_anchor=(0.5, 1.2), loc='upper center', ncol=3, fontsize='large' )
  plt.xticks(rotation=30)
  plt.ylabel('Reads number',fontsize=12)

  for x1, y1, y2, y3 in zip(data['steps'], x_have, x_not, x_failed):
    p1 = y1/(y1+y2+y3)
    p2 = y2/(y1+y2+y3)
    p3 = y3/(y1+y2+y3)
    # print( p1, p2, p3)
    if p1 >0.05:
      plt.text(x1, y1 * 0.8, '{:.0%}'.format(p1), ha='center', va="top")
    if p2>0.05:
      plt.text(x1, y1 + (y2)* 0.4,  '{:.0%}'.format(p2), ha='center')
    if p3>0.05:
      plt.text(x1, y1 + y2 + (y3)* 0.2, '{:.0%}'.format(p3), ha='center', va="bottom")

  # plt.yscale('log')
  formatter = ticker.ScalarFormatter(useMathText=True)
  formatter.set_scientific(True)
  formatter.set_powerlimits((-1,1))
  ax.yaxis.set_major_formatter(formatter)
  plt.tight_layout()
  plt.savefig(output_root +'imgs/reads_ratio.png')
  plt.savefig(output_root +'imgs/reads_ratio.pdf')
  plt.close(fig)
# {'steps': steps
# 'stackbar_list': stackbar_list}
fr = open(output_root + 'step_stackbar.pickle', 'rb')
step_stackbar = pickle.load(fr)
preprocess_summary_stack( step_stackbar, output_root)

# x->mapping tools
# y->uni, multi, unmap
def mapping_ratio_stack(data, output_root):
  x_un, x_uni, x_mul = data['map_ratio_list']
  fig = plt.figure(figsize=(8,8)) #设置画布的尺寸
  plt.bar(data['soft'], x_un, label="unmapped reads",edgecolor = 'black')
  plt.bar(data['soft'], x_uni, label="unique mapped reads",edgecolor = 'black', bottom = x_un)
  plt.bar(data['soft'], x_mul, label="non-unique mapped reads",edgecolor = 'black', bottom = [i+j for i, j in zip(x_un,x_uni)])
  # plt.legend( loc=3,fontsize=14)  # 设置图例位置
  plt.legend( bbox_to_anchor=(0.5, 1.2), loc='upper center', ncol=3, fontsize='large' )
  plt.ylabel('Mapping reads number',fontsize=16)
  plt.xlabel('Mapping reads split method',fontsize=16)

  for x1, y1, y2, y3 in zip(data['soft'], x_un, x_uni, x_mul):
    p1 = y1/(y1+y2+y3)
    p2 = y2/(y1+y2+y3)
    p3 = y3/(y1+y2+y3)
    if p1 >0.05:
      plt.text(x1, y1 * 0.4, '{:.0%}'.format(p1), ha='center',fontsize = 15)
    if p2>0.05:
      plt.text(x1, y1 + (y2)* 0.4,  '{:.0%}'.format(p2), ha='center',fontsize = 15)
    if p3>0.05:
      plt.text(x1, y1 + y2 + (y3)* 0.4, '{:.0%}'.format(p3), ha='center',fontsize = 15)

  plt.tight_layout()
  plt.savefig(output_root +'imgs/mapping_split.png')
  plt.savefig(output_root +'imgs/mapping_split.pdf')
  plt.close(fig)

mapping_ratio_data = {'map_ratio_list': [
    [0.15, 0.28 ],
    [0.83, 0.7],
    [0.02, 0.02]
],
'soft': ['bwa-mem', 'bowtie2']
}
# {'map_ratio_list': map_ratio_list
# 'soft': soft_list}
mapping_ratio_stack(mapping_ratio_data, output_root)

# todo replace boxplot to chr_rpkm_boxplot
# x-> chrs
# y-> boxs
def chr_rpkm_boxplot( data, output_root ):
  fig, ax = plt.subplots(figsize=(4,3))
  bp = ax.boxplot(data['rpkm_list'], patch_artist=True) #, boxprops={'facecolor': fixed_color_list[0], 'color':fixed_color_list[0] })

  # for box in bp['boxes']:
  #   box.set(facecolor='#087E8B', alpha=0.6, linewidth=2)

  for whisker in bp['whiskers']:
    whisker.set(linewidth=2)

  for median in bp['medians']:
    median.set(color='black', linewidth=3)

  # ax.set_title( data['title'] )
  # ax.set_xlabel( x_label)
  ax.set_ylabel( 'log10(RPKM)' )

  ax.set_xticklabels( [x.replace('chr', '') for x in data['sort_chr_list']], fontsize=7)
  fig.savefig(output_root + 'imgs/chr_rpkm.png')
  fig.savefig(output_root + 'imgs/chr_rpkm.pdf')
  # plt.yscale('log')
  plt.tight_layout()
  plt.close(fig)

# {'rpkm_list': rpkm_list,
# 'title': title,
# 'x_label': x_label,
# 'y_label': y_label,
# 'xtick_labels': xtick_labels}

fr = open(output_root + 'chr_rpkm.pickle', 'rb')
chr_rpkm= pickle.load(fr)
chr_rpkm_boxplot( chr_rpkm, output_root )

# big_size, big_label
# small_size, small_label
def expressed_pie(data, output_root):
  fig = plt.figure(figsize=(12, 7))
  # bigger
  bigger = plt.pie(data['big_sizes'], labels=data['big_labels'], labeldistance=0.75, startangle=90, frame=True, textprops={'fontsize': 12, 'color': 'w'})
  # [autotext.set_color('white') for autotext in autotexts]
  '''
  smaller = plt.pie(sizes_vegefruit, labels=labels_vegefruit,
                    colors=colors_vegefruit, radius=0.7,
                    startangle=90, labeldistance=0.7)
  '''

  wedges, texts = plt.pie(data['small_sizes'], wedgeprops=dict(width=0.5),
    radius=0.7, labeldistance=0.7,  startangle=90)
  bbox_props = dict(boxstyle="square,pad=0.3", fc="white", ec="k", lw=0.72)
  kw = dict(arrowprops=dict(arrowstyle="-", color='black'),
    bbox=bbox_props, zorder=100, va="center")

  for i, p in enumerate(wedges):
    ang = (p.theta2 - p.theta1)/2. + p.theta1
    y = np.sin(np.deg2rad(ang))
    x = np.cos(np.deg2rad(ang))

    horizontalalignment = {-1: "right", 1: "left"}[int(np.sign(x))]
    connectionstyle = "angle,angleA=0,angleB={}".format(ang)
    kw["arrowprops"].update({"connectionstyle": connectionstyle})
    plt.annotate(data['small_labels'][i],
    xy=(x*0.6, y*0.6), xytext=(1.35*np.sign(x), 1.4*y),
      horizontalalignment=horizontalalignment, **kw)

  centre_circle = plt.Circle((0, 0), 0.4, color='white', linewidth=0)
  # fig = plt.gcf()
  fig.gca().add_artist(centre_circle)

  plt.axis('equal')
  plt.tight_layout()
  plt.savefig(output_root + 'imgs/expressed_genes.png')
  plt.close(fig)


protein_num, linc_num = 18628, 5194
big_labels = [f'Coding\n({protein_num})', f'LncRNA\n({linc_num})']
expressed_protein_num, expressed_linc_num = 12000, 1500
expressed_protein_ratio = np.round( expressed_protein_num/protein_num, 2) * 100
expressed_linc_ratio = np.round( expressed_linc_num/linc_num, 2) * 100
big_sizes = [protein_num, linc_num]

small_labels = [f'{expressed_protein_ratio}% Protein-coding genes\nExpressed',
  f'{100-expressed_protein_ratio}% Protein-coding genes\nNot Expressed',
  f'{expressed_linc_ratio}% LncRNA genes\nExpressed',
  f'{100-expressed_linc_ratio}% LncRNA genes\nNot Expressed'
]
small_sizes = [expressed_protein_num, protein_num-expressed_protein_num, expressed_linc_num, linc_num - expressed_linc_num]
# { 'big_sizes': big_sizes,
# 'big_labels': big_labels,
# 'small_sizes': small_sizes,
# 'small_labels': small_labels}
expressed_pie_data = { 'big_sizes': big_sizes,
'big_labels': big_labels,
'small_sizes': small_sizes,
'small_labels': small_labels}
expressed_pie(expressed_pie_data, output_root)

def expressed_distribution(protein_x,protein_density,linc_x, linc_density, output_root):
  fig, ax = plt.subplots(figsize=(8,8))
  plt.plot(protein_x,protein_density(protein_x), label='Coding')
  plt.plot(linc_x,linc_density(linc_x), label='LncRNA')
  ax.set_xlabel('log10(RPKM)')
  ax.set_ylabel('density')
  ax.legend(loc='best')
  plt.tight_layout()
  plt.savefig(output_root + 'imgs/expression_distribution.png')
  plt.savefig(output_root + 'imgs/expression_distribution.pdf')
  plt.close(fig)

# fr = open(output_root + 'protein_linc_distribution.pickle', 'rb')
# data3 = pickle.load(fr)
# expressed_distribution( data3['protein_x'], data3['protein_density'],
# data3['linc_x'], data3['linc_density'], output_root)
# expressed_distribution()

# x->exon_intron_df.exon
# y->exon_intron_df.intron
# text
def exon_intron_scatter( data, output_root ):
  fig, ax = plt.subplots(figsize=(8,8))
  sns.regplot( data=data['filter_exon_intron_df1'], x='exon', y='intron', marker="x",
  scatter_kws = {'alpha' : 0.2 },
  line_kws={"color": 'black'}, x_jitter = 0.5, y_jitter = 0.5,  ax=ax, truncate = True )
  plt.text(0.2, 0.75, data['text'], transform = ax.transAxes)
  # plt.title( 'Exon/Intron ratio',  )
  plt.xlabel('Exon density')
  plt.ylabel('Intron density')
  plt.xlim([0, None])
  plt.ylim([0, None])

  fig.savefig(output_root + 'imgs/exon_intron_ratio.png')
  fig.savefig(output_root + 'imgs/exon_intron_ratio.pdf')
  fig.tight_layout()
  plt.close(fig)

fr = open(output_root + 'exonIntron.pickle', 'rb')
exon_intron_data = pickle.load(fr)
exon_intron_scatter( exon_intron_data, output_root )

def proximal_pausing_sites_distribution(site_index_list, output_root):
  fig, ax = plt.subplots(figsize=(12,7))
  sns.displot(data=pd.DataFrame({'pause site': site_index_list}), x='pause site', kde=True, ax=ax)
  plt.xlabel('Pause site')
  plt.ylabel('Frequency')
  plt.text(325, int(np.max(site_index_list)*0.9), s='Mean=%s\nMedian=%s'%(int(np.mean(site_index_list)), int(np.median(site_index_list))), verticalalignment="top",horizontalalignment="right")
  # plt.show()
  plt.savefig(output_root+'imgs/proximal_pause_site.png')
  plt.close(fig)

fr = open(output_root + 'proximal_pausing_sites.pickle', 'rb')
data5 = pickle.load(fr)
proximal_pausing_sites_distribution(data5['site_index_list'], output_root)



chr_strands = set([ln.strip().split('\t')[-1] for ln in open(output_root + 'pausing_sites.bed', 'a+') ] )
def sort_chr_list( chr_list ):
  sort_chr_list = [chrom for chrom in chr_list if re.match( r'chr\d+$', chrom)]
  sort_chr_list.sort(key=lambda arr: (arr[:3], int(arr[3:])))
  if 'chrM' in chr_list:  sort_chr_list.append('chrM')
  if 'chrMT' in chr_list:  sort_chr_list.append('chrMT')
  if 'chrX' in chr_list:  sort_chr_list.append('chrX')
  if 'chrY' in chr_list:  sort_chr_list.append('chrY')
  return sort_chr_list

sort_chrs = sort_chr_list( list(set([c[:-1] for c in chr_strands]) ) )
# chr_site_count_dic = defaultdict( {'site': [], 'count': []})
filter_chr_site_count_dic = {}
for ln in open(output_root + 'pausing_sites.bed', 'a+'):
  ls = ln.split('\t')
  chrom = ls[0]
  if chrom not in filter_chr_site_count_dic.keys():
    filter_chr_site_count_dic[chrom] = {'forward': {'site': [], 'count': []}, 'reverse': {'site': [], 'count': []}}
  if ls[-1].strip() == '+':
    filter_chr_site_count_dic[chrom]['forward']['site'].append(int(ls[1]) )
  else:
    filter_chr_site_count_dic[chrom]['reverse']['site'].append(int(ls[2]) )
  filter_chr_site_count_dic[chrom]['forward']['count'].append( int(ls[4]) )

print( filter_chr_site_count_dic )
def lollipopplot(chrs, lollipop_dic, output_file):
  fig, axes = plt.subplots(len(chrs)*2, 1, figsize=(  30, 2*len(chrs) ),  sharex=True, sharey=False)
  plt.xticks([])

  i = 0
  for chr in chrs:
    try:
      axes[i].stem( lollipop_dic[chr]['forward']['site'], lollipop_dic[chr]['forward']['count'], linefmt=color, markerfmt='.', basefmt=color  )
    except:
      pass
    axes[i].set_ylabel( chr+':+') #, color=color )
    axes[i].yaxis.tick_right()
    axes[i].yaxis.get_major_ticks()[1].set_visible(False)
    axes[i].yaxis.get_major_ticks()[-2].set_visible(False)
    axes[i].spines['top'].set_visible(False)
    axes[i].spines['left'].set_visible(False)
    # axes[i].spines['right'].set_color( color )
    i+=1
    try:
      axes[i].stem( lollipop_dic[chr]['reverse']['site'], [-1*x for x in lollipop_dic[chr]['reverse']['count']], linefmt=color, markerfmt='.', basefmt=color )
    except:
      pass
    axes[i].set_ylabel( chr+':-') # , color=color)
    axes[i].yaxis.tick_right()
    axes[i].yaxis.get_major_ticks()[1].set_visible(False)
    axes[i].yaxis.get_major_ticks()[-2].set_visible(False)
    axes[i].spines['bottom'].set_visible(False)
    axes[i].spines['left'].set_visible(False)
    # axes[i].spines['right'].set_color(color)
    i+=1

  # plt.subplots_adjust(left=0.1, right=0.9, top=0.9, bottom=0.1)
  # plt.title('Genome wide pausing sites distribution', y=1.1, fontsize=18)
  fig.tight_layout()
  fig.savefig(output_file, bbox_inches='tight')
  # plt.show()
  plt.close(fig)
lollipopplot(sort_chrs, filter_chr_site_count_dic, output_root + 'imgs/pause_sites.png')



import os, fire
# 输入 sam, 输出 unique_sam
from plot import mapping_ratio_stack

def parse_tags(s):
  tags = {}
  for tag in s.split('\t'):
    name, type_, val = tag.split(':')
    tags[name] = int(val) if type_ == 'i' else val
  return tags


def main(sam_dir, sam_file, output_root):
  # unmapped: SAM flag includes 4
  # 简单来说: 区分unique和nonredundant，nonredundant是绝对的单一比对，而unique是可以有多比对，但最高分就一个。
  # AS是比对的最高分，XS是第二高分。
  # uniquemapped: include multihit but uniquely determinable reads by alignment score
  # multimapped: multiple aligned reads with same scores
  # nonredundant: aligned exactly 1 time (bowtie2)
  # redundant: aligned >1 times(bowtie2), include uniquely determinable reads by alignment score
  total,unmapped, uniquemapped, multimapped, nonredundant, redundant  = 0, 0, 0, 0, 0, 0
  unmapped_sam = open(sam_dir + 'unmapped.sam', 'w')
  uniquemapped_sam = open(sam_dir + 'uniquemapped.sam', 'w')
  multimapped_sam = open(sam_dir + 'multimapped.sam', 'w')
  nonredundant_sam = open(sam_dir + 'nonredundant.sam', 'w')
  redundant_sam = open(sam_dir + 'redundant.sam', 'w')
  for ln in open(sam_dir + sam_file):
    if ln.startswith('@'):
      unmapped_sam.write(ln)
      uniquemapped_sam.write(ln)
      multimapped_sam.write(ln)
      nonredundant_sam.write(ln)
      redundant_sam.write(ln)
      continue
    _, flag, _, _, _, _, _, _, _, _, _, tags = ln.split('\t', 11)
    total += 1
    if int(flag) & 4:
      unmapped += 1
      unmapped_sam.write(ln)
    else:
      tags = parse_tags(tags)
      if "AS" not in tags.keys():
        continue
      as_ = tags["AS"]
      xs = tags.get("XS")
      if as_ != xs:
        uniquemapped += 1
        uniquemapped_sam.write(ln)
      else:
        multimapped += 1
        multimapped_sam.write(ln)
      if xs is None or (as_ != 0 and xs == 0):
        nonredundant += 1
        nonredundant_sam.write(ln)
      else:
        redundant += 1
        redundant_sam.write(ln)
  unmapped_sam.close()
  uniquemapped_sam.close()
  multimapped_sam.close()
  nonredundant_sam.close()
  redundant_sam.close()
  print('    unmapped--'+str(unmapped) )
  print('    uniquemapped--'+str(uniquemapped) )
  print('    multimapped--'+str(multimapped) )
  print('    nonredundant--'+str(nonredundant) )
  print('    redundant--'+str(redundant) )
  csv_dir = sam_dir.replace('/sam/', '/csv/mapping_report.csv')
  with open(csv_dir, 'w') as f:
    f.write( 'total_mapped,' + str(total)+'\n' )
    f.write( 'unmapped,'+ str(unmapped)+'\n' )
    f.write( 'unique_mapped,' + str(uniquemapped)+'\n' )
    f.write( 'multiple_mapped,' + str(multimapped)+'\n'  )
    f.write( 'nonredundant_mapped,' + str(nonredundant)+'\n')
    f.write( 'redundant_mapped,' + str(redundant)+'\n')

  stackbar_list = [
    [unmapped, unmapped ],
    [uniquemapped, nonredundant],
    [multimapped, redundant]
  ]
  mapping_ratio_data = {'map_ratio_list': stackbar_list,
  'soft': ['bwa-mem', 'bowtie2'] }
  mapping_ratio_stack(mapping_ratio_data, output_root)

if __name__ == '__main__':
  fire.Fire( main )



import os, re
import fire


def filter_config(output_root, config_list):
  ## 用tuple(type, file, key) 表示
  ## a 元素 是否存在，('ele', file, key)
  ## b 文件存在。 ('file', file, key)
  para_dic = {}
  for config_tuple in config_list:
    if config_tuple[0] == 'para':
      file, para = config_tuple[1], config_tuple[2]
      value = ''
      if os.path.exists(output_root + file):
        for ln in open(output_root + file):
          if ln.startswith(para):
            try:
              value = ln.split(',')[1].strip()
            except:
              print('no', file, para)
              continue
      # if value:
        # filter_config_list.append( (para, value) )
      para_dic[para] =value
    else:
      file, para = config_tuple[1], config_tuple[2]
      if os.path.exists(output_root + file):
        # filter_config_list.append( (para , file) )
        para_dic[para] =file
      else:
        para_dic[para] =""
  return para_dic

def render_template(para_dic,  output_root):
  # 把statics拷贝进output文件夹，
  template_dir = os.path.split(os.path.realpath(__file__))[0]+'/templates/'
  os.system( 'cp -r %s %s'%(template_dir + 'static/', output_root+'static/'))

  # 修改html的data，把参数字典导入到vue的data中,然后写入output的index.html
  f = open(output_root + 'index.html', 'w')
  template = open( template_dir + 'template.html').read()
  res = template.replace("data:{'replace':'here'}", 'data: %s'%str(para_dic) )
  # print( res )
  f.write(res)
  f.close()

def dic2str(dic):
  s = '{ '
  for k, v in dic.items():
    if isinstance(v, str):
      s+= k
      s+=': "'
      s+= v
      s+='",'
    if isinstance(v, list):
      s+= k + ': ["' + '","'.join(str(x) for x in v) + '"]'
  s+= ' }'
  return s

def main(output_root='./tmp_output/', type='all', is_server=''):
  ## type参数 可选 ['all', 'server',
  ## 'accessment', 'quantification',
  ## 'pausing', 'network']

  # 思路：
  ## html中引入 vue, 从而实现模板, 用变量和 v-if=变量 决定是否渲染
  ## 对所有参数值一一进行判断。例如文件是否存在
  ## 所有参数有个默认值 '', 不会显示。检查成功，这个默认值变成具体值

  # 步骤：
  # 1 生成所有参数，并对每个参数进行判断
  # 2 根据传入的type参数，决定哪些模块需要被show
  # 3 对所有值进行筛选
  # 3 渲染模板

  type_dic = {
    ## 所有参数按照模块分类(现在看分类没什么意义)
    # tuple包含3个值
    ## 1 para或file或img
    ## 2 file地址
    ## 3 变量名
    'preprocess': [
      # 1.0 file info
      ('para', 'csv/preprocess_report.csv', 'Read1_name'),
      ('para', 'csv/preprocess_report.csv', 'Read2_name'),
      ('para', 'csv/preprocess_report.csv', 'Read1_num'),
      ('para', 'csv/preprocess_report.csv', 'Read2_num'),
      ('para', 'csv/preprocess_report.csv', 'Read1_size'),
      ('para', 'csv/preprocess_report.csv', 'Read2_size'),
      # 1.1 adapter ratio
      ('para', 'csv/preprocess_report.csv', 'Reads_with_adapter'),
      ('para', 'csv/preprocess_report.csv', 'Uninformative_adapter_reads'),
      ('para', 'csv/preprocess_report.csv', 'Pct_uninformative_adapter_reads'),
      ('para', 'csv/preprocess_report.csv', 'Peak_adapter_insertion_size'),
      ('para', 'csv/preprocess_report.csv', 'Adapter_loss_rate'),
      # 1.2 RNA intergrity
      ('para', 'csv/preprocess_report.csv', 'Degradation_ratio'),
      ('img', 'imgs/adapter_insertion_distribution.png', 'img_adapter_distribution'),
      ('img', 'imgs/reads_distribution.png', 'img_reads_distribution'),
      ('img', 'imgs/reads_ratio.png', 'img_reads_ratio'),
      # 1.3 library complexity
      ('para', 'csv/mapping_report.csv', 'NRF'),
      ('para', 'csv/mapping_report.csv', 'PBC1'),
      ('para', 'csv/mapping_report.csv', 'PBC2'),
      # 1.4 QC trend
      ('img', 'imgs/quality_trend.png', 'img_quality_trend'),
      # 1.5 Nascent RNA purity
      ('para', 'csv/mapping_report.csv', 'assign_mapped'),
      ('para', 'csv/mapping_report.csv', 'chrM_mapped'),

      # 1.6 Other preprocess metric
      ('para', 'csv/preprocess_report.csv', 'Trimmed_reads'),
      ('para', 'csv/preprocess_report.csv', 'Trim_loss_rate'),
      ('para', 'csv/preprocess_report.csv', 'Reads_with_polyX'),
      ('para', 'csv/preprocess_report.csv', 'Uninformative_polyX_reads'),
      ('img', 'imgs/mapping_split.png', 'img_mapping_split'),

      # 1.7 output files
      ('file', 'fastq/clean_read1.fq.gz', 'clean_read1'),
      ('file', 'fastq/clean_read2.fq.gz', 'clean_read2'),
      ('file', 'sam/original.sam', 'original_sam'),
      ('file', 'sam/unmapped.sam', 'unmapped_sam'),
      ('file', 'sam/assign.sam', 'assign_sam'),
      ('file', 'sam/unassign.sam', 'unassign_sam'),
      ('file', 'sam/uniquemapped.sam', 'uniquemapped_sam'),
      ('file', 'sam/multimapped.sam', 'multimapped_sam'),
      ('file', 'sam/redundant.sam', 'redundant_sam'),
      ('file', 'sam/nonredundant.sam', 'nonredundant_sam'),
      ('file', 'bw/forward_bw', 'forward_bw'),
      ('file', 'bw/reverse_bw', 'reverse_bw'),

    ],
    'featureAssign': [
      # 2.1 Quantification & Normalization
      ('file', 'csv/all_feature_attrs.csv', 'all_feature_attrs_csv'),
      ('file', 'csv/lincRNA_baseCount.csv', 'lincRNA_baseCount_csv'),
      ('file', 'csv/lincRNA_gb_count.csv', 'lincRNA_gb_count_csv'),
      ('file', 'csv/lincRNA_pp_count.csv', 'lincRNA_pp_count_csv'),
      ('file', 'csv/lincRNA_rpkm.csv', 'lincRNA_rpkm_csv'),
      ('file', 'csv/lincRNA_rpm.csv', 'lincRNA_rpm_csv'),
      ('file', 'csv/protein_coding_baseCount.csv', 'protein_coding_baseCount_csv'),
      ('file', 'csv/protein_coding_gb_count.csv', 'protein_coding_gb_count_csv'),
      ('file', 'csv/protein_coding_pp_count.csv', 'protein_coding_pp_count_csv'),
      ('file', 'csv/protein_coding_rpkm.csv', 'protein_coding_rpkm_csv'),
      ('file', 'csv/protein_coding_rpkm.csv', 'protein_coding_rpm_csv'),
      ('file', 'csv/erna_quant.csv', 'erna_quant_csv'),
      # 3.1 PI $EI
      ('file', 'csv/lincRNA_ei.csv', 'lincRNA_ei_csv'),
      ('file', 'csv/lincRNA_pi.csv', 'lincRNA_pi_csv'),
      ('file', 'csv/protein_coding_ei.csv', 'protein_coding_ei_csv'),
      ('file', 'csv/protein_coding_pi.csv', 'protein_coding_pi_csv'),
      # 1.5
      ('file', 'csv/exon_intron_ratio.csv', 'exon_intron_ratio_csv'),
      ('img', 'imgs/exon_intron_ratio.png', 'img_exon_intron_ratio'),
      ('para', 'csv/preprocess_report.csv', 'mRNA_contamination'),
      # 2.2 Abundance esitimate
      ('img', 'imgs/expressed_genes.png', 'img_expressed_genes'),
      ('img', 'imgs/expression_distribution.png', 'img_expression_distribution'),
      # 2.3 Gene expression distribution
      ('img', 'imgs/chr_rpkm.png', 'img_chr_rpkm'),
      # 3.2 proximal pausing sites
      ('file', 'bed/proximal_pause_site.bed', 'proximal_pausing_site_bed'),
      ('img', 'imgs/proximal_pause_site.png', 'img_proximal_pause_site')
    ],
    'pausing_sites': [
      # 3.3 Global pausing sites
      ('file', 'bed/pausing_sites.bed', 'pausing_sites_bed'),
      ('img', 'imgs/pause_sites.png', 'img_pause_sites')
    ],
    'network': [
      # 4.1 Network measurement
      ('para', 'csv/graph_info.csv', 'nodes_num'),
      ('para', 'csv/graph_info.csv', 'edges_num'),
      ('para', 'csv/graph_info.csv', 'mean_degree'),
      ('para', 'csv/graph_info.csv', 'assortativity_coefficient'),
      ('para', 'csv/graph_info.csv', 'correlation_coefficient'),
      ('para', 'csv/graph_info.csv', 'transitivity'),
      ('para', 'csv/graph_info.csv', 'density'),
      # 4.2 Degree distribution
      ('img', 'imgs/network_degree.png', 'img_network_degree'),
      # 4.3 motif distribution
      ('img', 'imgs/network_motif.png', 'img_network_motif')
    ]
  }
  if not output_root.endswith('/'): output_root = output_root +'/'
  config_list = type_dic['preprocess'] + type_dic['featureAssign'] + type_dic['pausing_sites'] + type_dic['network']
  para_dic = filter_config(output_root, config_list)

  ## 渲染模块的字典
  type_module_dic = {
    'server': {'show_basic': '', 'show_assess': 'true', 'show_quant': 'true', 'show_pause': 'true', 'show_network': 'true'},
    'all': {'show_basic':'true', 'show_assess': 'true', 'show_quant': 'true', 'show_pause': 'true', 'show_network': 'true'},
    'assessment': {'show_basic':'', 'show_assess': 'true', 'show_quant': '', 'show_pause': '', 'show_network': ''},
    'quantification': {'show_basic':'', 'show_assess': '', 'show_quant': 'true', 'show_pause': '', 'show_network': ''},
    'pausing': {'show_basic':'', 'show_assess': '', 'show_quant': '', 'show_pause': 'true', 'show_network': ''},
    'network': {'show_basic':'', 'show_assess': '', 'show_quant': '', 'show_pause': '', 'show_network': 'true'},
  }
  show_module_dic = type_module_dic[type]
  para_dic.update( show_module_dic )

  para_dic['is_server'] = is_server



  ## 单独 为 network模块 识别一下 communities
  if para_dic['show_network']:
    community_list = []
    for community_img in os.listdir(output_root +'imgs/'):
      if community_img.startswith('community_'):
        num = int( community_img.split('_')[1].replace('.png', '') )
        community_list.append( num )
    if community_list == []:
      para_dic['community_list'] = ''
    else:
      # 4.4 community detection
      para_dic['community_list'] = ['community_%s'%str(x) for x in sorted(community_list)]

  para_str = dic2str( para_dic )
  # print(para_str)
  # 定义好参数 给模板的默认值，大多数为None
  render_template(para_str, output_root)


if __name__ == '__main__':
  fire.Fire( main )


import os, sys
script_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), ".."))
lib_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), '../../libs') )
sys.path.append(lib_dir)
output_root = sys.argv[1]
if not output_root.endswith('/'): output_root = output_root +'/'
from collections import Counter

import pandas as pd
import numpy as np

from py_ext import json2dic, get_file_size
from plot import insertion_distribution, readsLength_distribution, q20_q30_line, preprocess_summary_stack

# 目标:
# 1 生成 preprocess_report.txt
# 2 生成 3张图片

global stats_list
stats_list = []

def parse_fastp_json(json_dir):
  fastp_dic = json2dic(json_dir)
  summary = fastp_dic['summary']
  before = summary['before_filtering']
  after = summary['after_filtering']
  res = fastp_dic['filtering_result']

  para_dic = {
   'before_reads': int(before['total_reads']),
   'before_bases': int(before['total_bases']),
   'after_reads': int(after['total_reads']),
   'after_bases': int(after['total_bases']),

   'before_q20_rate': float(before['q20_rate']),
   'before_q30_rate': float(before['q30_rate']),
   'before_read1_mean_length': float(before['read1_mean_length']),
   'before_gc_rate': float(before['gc_content']),
   'after_q20_rate': float(after['q20_rate']),
   'after_q30_rate': float(after['q30_rate']),
   'after_reads_mean_length': float(after['read1_mean_length']),
   'after_gc_rate': float(after['gc_content']),

   'passed_filter_reads': int(res['passed_filter_reads']),
   'low_quality_reads': int(res['low_quality_reads']),
   'too_short_reads': int(res['too_short_reads']),
   'too_many_N_reads': int(res['too_many_N_reads'])
  }
  if 'read2_mean_length' in before.keys():
    para_dic['before_read2_mean_length'] = float(before['read2_mean_length'])
  if 'read2_mean_length' in after.keys():
    para_dic['after_read2_mean_length'] = float(after['read2_mean_length'])
  if 'adapter_cutting' in fastp_dic.keys():
    para_dic['adapter_trimmed_reads'] = fastp_dic['adapter_cutting']['adapter_trimmed_reads']
  if 'polyx_trimming' in fastp_dic.keys():
    para_dic['polyx_trimmed_reads'] = fastp_dic['polyx_trimming']['total_polyx_trimmed_reads']
  return para_dic


# 0 get variable
tmp_variable_dic = {ln.split('--')[0]: ln.split('--')[1].strip() for ln in open(output_root+"tmp_variable.txt")}
# stats_list.append( ['Read1_name', os.path.basename(tmp_variable_dic['read1_name'])] )
# stats_list.append( ['Read1_size',get_file_size(tmp_variable_dic['read1_name'])] )
stats_list.append( ['Read1_num', tmp_variable_dic['read1_num']] )

try:
  # stats_list.append( ['Read2_name', os.path.basename(tmp_variable_dic['read2_name'])] )
  # stats_list.append( ['Read2_size',get_file_size(tmp_variable_dic['read2_name'])] )
  stats_list.append( ['Read2_num', tmp_variable_dic['read2_num']] )
except:
  pass

stats_list.append( ['Reads_with_adapter', tmp_variable_dic['reads_with_adapter']] )
stats_list.append( ['Uninformative_adapter_reads', tmp_variable_dic['uninformative_adapter_reads']] )
stats_list.append( ['Pct_uninformative_adapter_reads', tmp_variable_dic['pct_uninformative_adapter_reads']] )
stats_list.append( ['Peak_adapter_insertion_size', tmp_variable_dic['peak_adapter_insertion_size'] ] )

remove_adapter_dic = parse_fastp_json(output_root+"json/remove_adapter.json")
cut_twoEnd_dic = parse_fastp_json(output_root+"json/trim.json")
filter_quality_dic = parse_fastp_json(output_root+"json/filter_quality.json")
remove_polyX_dic = parse_fastp_json(output_root+"json/remove_polyX.json")

q20_list, q30_list = [], []
# 1.1 raw reads
total_reads = remove_adapter_dic['before_reads']
s1_1 = total_reads
s1_2, s1_3 = 0, 0

table_data = []

annotate_text1 = 'Total reads: {reads}\nTotal bases: {bases}\nMean length: {mean_length}\nGC rate: {gc}\nQ20: {q20}\nQ30: {q30}'.format(
  reads=total_reads, bases=remove_adapter_dic['before_bases'], mean_length=remove_adapter_dic['before_read1_mean_length'],
  q20=remove_adapter_dic['before_q20_rate'], q30=remove_adapter_dic['before_q30_rate'], gc=remove_adapter_dic['before_gc_rate']
)
q20_list.append( remove_adapter_dic['before_q20_rate'] )
q30_list.append( remove_adapter_dic['before_q30_rate'] )
table_data.append( [total_reads, remove_adapter_dic['before_bases'], remove_adapter_dic['before_read1_mean_length'],
  remove_adapter_dic['before_gc_rate'], remove_adapter_dic['before_q20_rate'], remove_adapter_dic['before_q30_rate']]
)

# 1.2 remove adapter
pass_adapter = remove_adapter_dic['passed_filter_reads']
fail_adapter = remove_adapter_dic['too_short_reads']
with_adapter = remove_adapter_dic['adapter_trimmed_reads']
s2_1 = with_adapter
s2_2 = pass_adapter - with_adapter
s2_3 = fail_adapter
annotate_text2 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_adapter, bases=remove_adapter_dic['after_bases'], mean_length=remove_adapter_dic['after_reads_mean_length'],
  q20=remove_adapter_dic['after_q20_rate'], q30=remove_adapter_dic['after_q30_rate'], gc=remove_adapter_dic['after_gc_rate']
)
q20_list.append(remove_adapter_dic['after_q20_rate'] )
q30_list.append(remove_adapter_dic['after_q30_rate'] )
table_data.append([pass_adapter, remove_adapter_dic['after_bases'], remove_adapter_dic['after_reads_mean_length'],
  remove_adapter_dic['after_gc_rate'], remove_adapter_dic['after_q20_rate'], remove_adapter_dic['after_q30_rate']
])
stats_list.append( ['Reads_with_adapter', with_adapter] )
stats_list.append( ['Uninformative_adapter_reads', fail_adapter] )
stats_list.append( ['Pct_uninformative_adapter_reads', round(fail_adapter/(pass_adapter+fail_adapter)*100, 3)] )
# stats_list.append( ['Peak_adapter_insertion_size', peak_adapter_insertion_size] )
before_adapter_bases = remove_adapter_dic['before_bases']
after_adapter_bases = remove_adapter_dic['after_bases']
adapter_loss_rate = (before_adapter_bases - after_adapter_bases)/before_adapter_bases
stats_list.append( ['Adapter_loss_rate', adapter_loss_rate] )

insertion_size_list = [int( ln.split('\t')[1].strip() ) for ln in open(output_root + 'txt/adapter_read_len1.txt')]
# 计算 y->insertion size的频率(Counter就行)
insert_size_counter = Counter(insertion_size_list)
x_list = list( insert_size_counter.keys() )
y_list = list( insert_size_counter.values() )
x_max = max(x_list)  # x_max对应reads 没有去掉 adapter
insertion_length, length_count = [], []
degraded_reads, intact_reads = 0, 0
for x, y in zip(x_list, y_list):
  if x==x_max:
    continue
  l = x_max - x
  insertion_length.append( l )
  length_count.append(y)
  if x >= 10 and x < 20:
    degraded_reads+=y
  if x >= 30 and x < 40:
    intact_reads+=y

degradation_ratio = degraded_reads/float(intact_reads)
stats_list.append( ['Degradation_ratio', degradation_ratio] )

insertion_dist_data =  {
  'insertion_length': insertion_length,
  'length_count': length_count,
  'degradation_ratio': degradation_ratio,
}

insertion_distribution( insertion_dist_data, output_root  )
# with open('insertion_dist_data.pickle', 'wb') as f:
#   pickle.dump(insertion_dist_data, f)

# 1.3 trim two ends
pass_twoEnd = cut_twoEnd_dic['passed_filter_reads']
fail_twoEnd = cut_twoEnd_dic['too_short_reads']
with_twoEnd = int(tmp_variable_dic['reads_with_cutTwoEnd'])
s3_1 = with_twoEnd
s3_2 = pass_twoEnd - with_twoEnd
s3_3 = fail_twoEnd
annotate_text3 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_twoEnd, bases=cut_twoEnd_dic['after_bases'], mean_length=cut_twoEnd_dic['after_reads_mean_length'],
  q20=cut_twoEnd_dic['after_q20_rate'], q30=cut_twoEnd_dic['after_q30_rate'], gc=cut_twoEnd_dic['after_gc_rate']
)
q20_list.append( cut_twoEnd_dic['after_q20_rate'] )
q30_list.append( cut_twoEnd_dic['after_q30_rate'] )
table_data.append([ pass_twoEnd, cut_twoEnd_dic['after_bases'], cut_twoEnd_dic['after_reads_mean_length'],
  cut_twoEnd_dic['after_gc_rate'], cut_twoEnd_dic['after_q20_rate'], cut_twoEnd_dic['after_q30_rate']
])

stats_list.append( ['Trimmed_reads', with_twoEnd ])
before_trim_bases = cut_twoEnd_dic['before_bases']
after_trim_bases = cut_twoEnd_dic['after_bases']
trim_loss_rate= round( (before_trim_bases-after_trim_bases)/ before_trim_bases*100, 2)
stats_list.append( ['Trim_loss_rate', trim_loss_rate] )

# 1.4 remove polyX
pass_polyX = remove_polyX_dic['passed_filter_reads']
fail_polyX = remove_polyX_dic['too_short_reads']
with_polyX = remove_polyX_dic['polyx_trimmed_reads']
s4_1 = with_polyX
s4_2 = pass_polyX - with_polyX
s4_3 = fail_polyX
annotate_text4 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_polyX, bases=remove_polyX_dic['after_bases'], mean_length=remove_polyX_dic['after_reads_mean_length'],
  q20=remove_polyX_dic['after_q20_rate'], q30=remove_polyX_dic['after_q30_rate'], gc=remove_polyX_dic['after_gc_rate']
)
q20_list.append( remove_polyX_dic['after_q20_rate'] )
q30_list.append( remove_polyX_dic['after_q30_rate'] )
table_data.append([pass_polyX, remove_polyX_dic['after_bases'], remove_polyX_dic['after_reads_mean_length'],
  remove_polyX_dic['after_gc_rate'], remove_polyX_dic['after_q20_rate'], remove_polyX_dic['after_q30_rate']
])
stats_list.append( ['Reads_with_polyX', with_polyX] )
stats_list.append( ['Uninformative_polyX_reads', fail_polyX] )

# 1.5 mean quality
pass_quality = filter_quality_dic['passed_filter_reads']
fail_quality = filter_quality_dic['low_quality_reads'] + filter_quality_dic['too_many_N_reads']
s5_1 = pass_quality
s5_2 = 0
s5_3 = fail_quality
annotate_text5 = 'Total reads: {reads}\nTotal bases: {bases}\nReads mean length: {mean_length}\nQ20 rate: {q20}\nQ30 rate: {q30}\nGC rate: {gc}'.format(
  reads=pass_quality, bases=filter_quality_dic['after_bases'], mean_length=filter_quality_dic['after_reads_mean_length'],
  q20=filter_quality_dic['after_q20_rate'], q30=filter_quality_dic['after_q30_rate'], gc=filter_quality_dic['after_gc_rate']
)
q20_list.append( filter_quality_dic['after_q20_rate'] )
q30_list.append( filter_quality_dic['after_q30_rate'] )
table_data.append([ pass_quality, filter_quality_dic['after_bases'], filter_quality_dic['after_reads_mean_length'],
  filter_quality_dic['after_gc_rate'], filter_quality_dic['after_q20_rate'], filter_quality_dic['after_q30_rate']
])

stackbar_list = [
  # have not failed
  [s1_1, s2_1, s3_1, s4_1, s5_1],
  [s1_2, s2_2, s3_2, s4_2, s5_2],
  [s1_3, s2_3, s3_3, s4_3, s5_3]
]
steps=['raw reads', 'adapter removed', 'trim', 'polyX removed', 'quality filtered']
step_stackbar = {'steps': steps, 'stackbar_list': stackbar_list}
preprocess_summary_stack(step_stackbar, output_root)
# with open('step_stackbar.pickle', 'wb') as f:
#   pickle.dump(step_stackbar, f)

q20_q30_data = {'q20_list': q20_list,
'q30_list': q30_list,
'steps': steps}
q20_q30_line( q20_q30_data, output_root)
# with open('q20_q30_data.pickle', 'wb') as f:
#   pickle.dump(q20_q30_data, f)

def reads_length_dist(reads_len_file, min, max):
  reads_length = [int(ln.split('\t')[1]) for ln in open(reads_len_file) ]
  length_counter = Counter(reads_length)
  x = list( range(min, max+1) )
  y = [length_counter[i] if i in length_counter.keys() else 0 for i in x ]
  return [x, y]

read_max_length = int( remove_adapter_dic['before_read1_mean_length'] )

remove_adapter_list = reads_length_dist(output_root + "/txt/adapter_read_len1.txt", 1, read_max_length)
cut_twoEnd_list = reads_length_dist(output_root + "/txt/filter_trim_len1.txt", 1, read_max_length)
remove_polyX_list = reads_length_dist(output_root + "/txt/filter_polyX_len1.txt", 1, read_max_length)
filter_quality_list = reads_length_dist(output_root + "/txt/filter_quality_len1.txt", 1, read_max_length)

bar_list = [[list(range(1, read_max_length+1)), [0]*(read_max_length-1) + [total_reads]], filter_quality_list, remove_adapter_list, remove_polyX_list, cut_twoEnd_list]
table_col = ['Total reads', 'Total bases', 'Mean length', 'GC rate', 'Q20', 'Q30']
table_row = ['raw', 'adapter removed', 'trimed', 'polyX removed', 'quality filtered']

reads_length_data = {'steps': steps, 'bar_list': bar_list,
'annotate_list': [annotate_text1, annotate_text2, annotate_text3, annotate_text4, annotate_text5]}
readsLength_distribution(reads_length_data, output_root)
# with open('reads_length_data.pickle', 'wb') as f:
#   pickle.dump(reads_length_data, f)

c = open(output_root + 'csv/preprocess_report.csv', 'w')
for stat_item in stats_list:
  c.write(stat_item[0] + ',' + str(stat_item[1]) + '\n')
c.close()



import os, re, json, requests, fire
import pandas as pd
from loguru import logger

# 思路:
# (提取参数，检验软件，检验参数 在nasap文件中做)
# 根据调控数据下载调控数据
# 生成模板

def parse_regulatory_data(chrom, start, end,  resp_data):
  source_target_gene_list, link_list = [], []

  for regulator_dic in resp_data["regulator"]:
    regulator_chr = regulator_dic['c']
    regulator_start = regulator_dic['s']
    regulator_end = regulator_dic['e']
    regulator_name = regulator_dic['n']

    for target_obj in regulator_dic['targets']:
      target_dic = target_obj["target_site"]
      target_chr = target_dic['c']
      target_start = target_dic['s']
      target_end = target_dic['e']
      target_name = target_dic['n']

      if target_start >= int(start) and target_end <= int(end):
        ln=str(regulator_chr) + '\t' + str(regulator_start) + '\t' + str(regulator_end) + '\t' + str(target_chr) + '\t' + str(target_start) + '\t' + str(target_end) + '\n'
        link_list.append(ln)
        source_target_gene_list.append( (regulator_name, target_name) )

  return link_list, source_target_gene_list


def generate_link_file( specie, regulation_region, regulation_type ):
  chrom, start, end = re.split(r'[:-]+', regulation_region)
  try:
    # 网路获取 enhancer调控数据
    # print( 'http://grobase.top/meta/api/rest/{regulation}/{specie}/{chr}/{start}/{end}'.format(regulation=regulation_type, specie=specie, chr=chrom, start=start, end=end) )
    resp = requests.get('http://grobase.top/meta/api/rest/{regulation}/{specie}/{chr}/{start}/{end}'.format(regulation=regulation_type, specie=specie, chr=chrom, start=start, end=end))
    resp_data = json.loads( resp.text )

    link_list, source_target_gene_list = parse_regulatory_data(chrom, start, end, resp_data )
    return [link_list, source_target_gene_list]
  except:
    print( 'request failed for %s data'%regulation_type)
    return ''


def render_track_template(para_dic, show_dic, gtf, output_root):
  template = open(os.path.split(os.path.realpath(__file__))[0] + '/templates/template_track.txt').read()

  f = open(output_root + 'txt/track.txt', 'w')
  # print( para_dic )
  for k, v in para_dic.items():
    # print('$'+k, v)
    template = template.replace( '$'+k, v )

  for k, v in show_dic.items():
    if v == False:
      regex = re.compile('#show\-%s[^"]*#'%k)
      template = re.sub(regex, '', template)

  f.write(template)
  f.close()

def main( specie, region, gtf, forward, reverse, output_root, rpkm_file='' ):
  show_dic = {
    'enhancer': False,
    'tf': False
  }
  if not output_root.endswith('/'): output_root = output_root +'/'

  specie_regulation_dic = {
    'arabidopsis': ['tf'],
    'chicken': ['enhancer'],
    'chimp': ['enhancer'],
    'fly': ['enhancer', 'tf'],
    'frog': ['enhancer'],
    'mouse': ['enhancer', 'tf'],
    'human': ['enhancer', 'tf'],
    'rat': ['enhancer', 'tf'],
    'rhesus': ['enhancer'],
    'sheep': ['enhancer'],
    'worm': ['enhancer', 'tf'],
    'yeast': ['tf'],
    'zebrafish': ['enhancer', 'tf']
  }

  if specie not in specie_regulation_dic.keys():
    print('no regulatory data for specie ' + specie)
    os.sys.exit(1)

  para_dic = { 'specie': specie, 'gtf': gtf, 'forward': forward, 'reverse': reverse }
  regulation_list = specie_regulation_dic[specie]

  # 表达为0的基因 被filter
  filter_gene_list = []
  # print( 'rpkm', rpkm_file)
  if rpkm_file:
    rpkm_df = pd.read_csv( rpkm_file, index_col = 0 )
    filter_gene_list.extend( list( rpkm_df[rpkm_df.rpkm ==0].index ) )
  print('filter genes', len(filter_gene_list))
  filter_n = 0
  for regulation_type in regulation_list:
    logger.info('genome_tracks_visualize--generate %s regulatory link file'%regulation_type)
    # 获取 调控区
    link_data = generate_link_file( specie, region, regulation_type )
    # print(link_data )
    if link_data != '':
      show_dic[regulation_type] = True
      link_list, source_target_gene_list = link_data

      with open(output_root + 'txt/%s_link.txt'%regulation_type, 'w' ) as f:
        for link_ln, g in zip(link_list, source_target_gene_list):
          if (g[0] not in filter_gene_list) and (g[1] not in filter_gene_list):
            f.write(link_ln)
          else:
            filter_n +=1

      para_dic[ regulation_type ] = os.path.abspath(output_root + 'txt/%s_link.txt'%regulation_type)

  print( 'filtered num is', filter_n)

  render_track_template(para_dic, show_dic, gtf, output_root)
  # 渲染图片
  logger.info('genome_tracks_visualize--start plotting genome tracks')
  os.system('pyGenomeTracks --tracks {track_dir} --region {region} -o {img_dir}'.format(track_dir=output_root + 'txt/track.txt', region=region, img_dir=output_root + 'imgs/genome_track.png' ) )
  os.system('pyGenomeTracks --tracks {track_dir} --region {region} -o {img_dir}'.format(track_dir=output_root + 'txt/track.txt', region=region, img_dir=output_root + 'imgs/genome_track.pdf' ) )
  logger.info('genome_tracks_visualize--your results can be found at %s'%output_root)




if __name__ == "__main__":
  fire.Fire( main )



import os, fire, pickle

import numpy as np
import pandas as pd
import pyBigWig
from collections import defaultdict

from parse_gtf import sort_chr_list
from plot import global_distribution_lollipopplot

def get_bw_pause_sites(para_list):
  [bw_dir, chr, size, isForward] = para_list
  win_side= 100 # 左右100，总共201
  bw = pyBigWig.open(bw_dir)

  start, end = 0, size
  gene_depth_array = bw.values(chr, start, end, numpy=True)
  # 获取小窗的window np arrays
  suit_index = np.where(gene_depth_array>=4)[0]
  suit_index = suit_index[suit_index >win_side]
  suit_index = suit_index[suit_index < end-win_side]

  site_list = []
  count_list = []

  for i in suit_index:
    window=gene_depth_array[i-win_side:i+win_side+1]
    win_count_max=window.max()
    if win_count_max <4: continue
    # 201bp小窗 中点
    if window[win_side] <4: continue
    if window[win_side] != win_count_max: continue

    win_count_mean=window.mean()
    # win_count_median=np.median( window )
    win_count_std=window.std()
    if win_count_max>=(win_count_mean + win_count_std*3):
      # return 1
      site_list.append(i)
      count_list.append( win_count_max)
    # return 0
  # suit_value = np.array( list( map(judge_window, suit_index) ) )
  # chr_sites = suit_index[suit_value >0]
  if isForward:
    chr_strand = chr + '+'
  else:
    chr_strand = chr + '-'
  # print( {chr_strand:  {'site': site_list, 'count': count_list} } )
  return {chr_strand:  {'site': site_list, 'count': count_list} }

# 思路:
# 1 获取所有染色体
# 2 单个染色体运行并生成结果
# 3 写入识别的暂停位点
# 4 画图
def main(forward_bw, reverse_bw, output_root='./tmp_output/', cores=1):
  filter200_chr_site_dic = {}
  for file_type, bw_dir in {'forward': forward_bw, 'reverse': reverse_bw}.items():
    isForward = True
    if file_type == 'reverse': isForward = False
    bw_value = pyBigWig.open(bw_dir)
    chr_sizes = {chr:size for chr, size in bw_value.chroms().items() }
    chr_size_list = list(chr_sizes.items())
    for chr_size_tuple in chr_size_list:
      # print( chr_size_tuple)
      # print( chr_size_tuple )
      chrom, size = chr_size_tuple[0], chr_size_tuple[1]
      chr_strand_site_dic = get_bw_pause_sites([bw_dir, chrom, size, isForward])
      # print( chr_strand_site_dic)

      for chr_strand, site_count_dic in chr_strand_site_dic.items():
        sort_site_list = site_count_dic['site']
        count_list = site_count_dic['count']

        final_index = len(sort_site_list) - 1
        if final_index == -1: continue
        if final_index == 0:
          filter200_chr_site_dic[chr_strand] = {'site': sort_site_list}
          filter200_chr_site_dic[chr_strand]['count'] =  count_list
          continue
        if final_index == 1:
          if (sort_site_list[1] - sort_site_list[0]) > 200:
            filter200_chr_site_dic[chr_strand] = {'site': sort_site_list}
            filter200_chr_site_dic[chr_strand]['count'] =  count_list
          continue

        tmp_list = []
        tmp_count_list = []
        last_index, cur_index, next_index =0, 1, 2
        if (sort_site_list[1] - sort_site_list[0]) >200:
          tmp_list.append( sort_site_list[0])
          tmp_count_list.append( count_list[0])

        while cur_index < final_index:
          last_value = sort_site_list[last_index]
          cur_value = sort_site_list[cur_index]
          next_value = sort_site_list[next_index]
          if (cur_value - last_value) > 200 and (next_value - cur_value) > 200:
            tmp_list.append(cur_value)
            tmp_count_list.append( count_list[cur_index] )
          last_index+=1
          cur_index+=1
          next_index+=1
        if (sort_site_list[final_index] - sort_site_list[final_index-1]) >200:
          tmp_list.append( sort_site_list[final_index])
          tmp_count_list.append( count_list[final_index])
        if len(tmp_list) == 0: continue
        filter200_chr_site_dic[chr_strand] = {'site': tmp_list, 'count': tmp_count_list}

  with open(output_root + 'bed/pausing_sites.bed', 'a+') as f:
    chr_strands = list(filter200_chr_site_dic.keys() )
    # print( list(set([c[:-1] for c in chr_strands]) ) )
    sort_chrs = sort_chr_list( list(set([c[:-1] for c in chr_strands]) ) )
    # print( sort_chrs )
    if not sort_chrs:
      print( 'Warning: no pausing sites to draw lollipop plot')
      os.sys.exit(1)
    filter_chr_site_count_dic  = defaultdict(lambda: {})
    for chr in sort_chrs:
      if chr+'+' in chr_strands:
        filter_chr_site_count_dic[chr]['forward'] = filter200_chr_site_dic[chr + '+']
      else:
        filter_chr_site_count_dic[chr]['forward'] = {'site': [], 'count': []}

      if chr+'-' in chr_strands:
        filter_chr_site_count_dic[chr]['reverse'] = filter200_chr_site_dic[chr + '-']
      else:
        filter_chr_site_count_dic[chr]['reverse'] = {'site': [], 'count': []}

    global_pause_sites_data = {'chrs': sort_chrs, 'lollipop_dic': filter_chr_site_count_dic}
    global_distribution_lollipopplot(global_pause_sites_data, output_root)
    # with open('global_pause_sites_data.pickle', 'wb') as f:
    #   pickle.dump(global_pause_sites_data, f)

    n =0
    for chr_strand, site_count_dic in filter200_chr_site_dic.items():
      site_list, count_list = site_count_dic['site'], site_count_dic['count']
      if len(site_list)==0: continue
      chr, strand = chr_strand[:-1], chr_strand[-1]
      for site, count in zip(site_list, count_list):
        f.write(chr+'\t'+ str(site) + '\t' + str(site+1) + '\t' + 'ps_'+str(n) + '\t'+ str(count) + '\t' + strand + '\n')
        n+=1

if __name__ == '__main__':
  fire.Fire( main )


import os, fire, re, sys, math
# sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))+ '/libs')
script_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), ".."))
lib_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), '../../libs') )
sys.path.append(lib_dir)
from collections import Counter

import numpy as np
import pandas as pd
from scipy import stats

import pyBigWig

from parse_gtf import get_gene_df, get_gene_merge_exon_dic, gene_df2dic
from plot import chr_rpkm_boxplot, expressed_pie, expressed_distribution, exon_intron_scatter, proximal_pausing_sites_distribution

## 这个模块可以单独使用
# nasap feature_assign --gtf=./a.gtf --forward=./a.bw --reverse=./b.bw --outpt=./tmp_out

def get_total_bases(forward_bw, reverse_bw):
  return pyBigWig.open(forward_bw).header()['sumData'] + pyBigWig.open(reverse_bw).header()['sumData']

def density2rpkm(density, total_bases):
  return density / total_bases * 1e6

def find_true_tss(chr, left, right, bw, strand):
  # ref: Defining the Status of RNA Polymerase at Promoters
  # Reannotation of TSSs
  # annotated TSS ± 150, highest number of reads within this window
  # If two bases both the same maximum, the most 5′-base was used as the TSS
  v_list = bw.values(chr, left, right)
  max_index_list = np.argwhere(v_list == np.amax(v_list)).flatten().tolist()
  if strand == '+':
    max_index = max_index_list[0]
  else:
    max_index = max_index_list[-1]
  return left + max_index


def search_truncated_region(arr):
  window_size = 100
  step_size = 25
  read_threshold = 25
  length_threshold = 800
  new_shape = ((arr.size - window_size) // step_size + 1, window_size)
  new_strides = (arr.strides[0] * step_size, arr.strides[0])
  arr_strided = np.lib.stride_tricks.as_strided(arr, shape=new_shape, strides=new_strides)
  arr_sum = arr_strided.sum(axis=1)
  read_num_arr = arr_sum/window_size

  # 获取大于threshold的元素的索引
  index = np.nonzero(read_num_arr > read_threshold)[0]
  # 计算索引的差值，找到连续索引的起点和长度
  ranges = np.split(index, np.where(np.diff(index) != 1)[0]+1)
  # print( ranges )
  res = []
  if ranges:
    for r in ranges:
        if len(r) > 0 and len(r) * step_size > length_threshold:
          res.append( [r[0]*step_size, len(r) * step_size] )
  return res



def pausing_index( chr, start, end, strand,  bw, tss_up, tss_down ):
  # 获取 表达的蛋白， lincRNA
  if strand == '+':
    true_tss = find_true_tss(chr, start-tss_up, start + tss_down, bw, strand)
    pp = bw.stats( chr, true_tss-tss_up, true_tss + tss_down)[0]
    gb = bw.stats( chr, true_tss +tss_down, end)[0]
  else:
    true_tss = find_true_tss(chr, end -tss_down, end + tss_up, bw, strand)
    pp = bw.stats( chr, true_tss -tss_down, true_tss + tss_up)[0]
    gb = bw.stats( chr, start, true_tss - tss_down)[0]
  ppl = tss_up + tss_down
  gbl = end - start - ppl
  pp_count, gb_count = int(pp * ppl ), int(gb* gbl )
  if gb == 0: gb = 1e-6
  # pi = float('nan')
  # pi= 'gene body count zero'
  pi = pp/gb
  return pi, pp_count, gb_count

def elongation_index( chr, start, end, strand,  bw ):
  # 获取 表达的蛋白， lincRNA
  if strand == '+':
    pp = bw.stats( chr, start - 100, start + 300)[0]
    gb = bw.stats( chr, start + 300, start + 2000)[0]
  else:
    pp = bw.stats( chr, end - 300, end + 100)[0]
    gb = bw.stats( chr, end - 2000, end - 300)[0]
  if pp == 0:
    #
    prr = 'proximal promoter count zero'
  else:
    prr = gb/pp
  return prr

def get_attrs(gene_range_dic, forward_bw, reverse_bw, total_bases, tss_up=0, tss_down=150, gene_length=2000, rpkm_threshold=0.1):
  gene_rpkm_dic, gene_baseCount_dic, gene_rpm_dic =  {},{}, {}
  gene_pi_dic, gene_ei_dic = {}, {}
  gene_pp_count_dic, gene_gb_count_dic = {}, {}
  trunc_gene_list = []
  for strand in ['+', '-']:
    strand_gene_range_dic = gene_range_dic[strand]
    bw = pyBigWig.open( {'+': forward_bw, '-': reverse_bw}[strand])
    for gene_name, gene_range in strand_gene_range_dic.items():
      c, s, e = gene_range[0], gene_range[1], gene_range[2]
      try:
        density = bw.stats(c, s, e)[0]
        rpkm = density2rpkm(density, total_bases)
        base_count = int( density * (e - s) )
        rpm = base_count * 1e6 /total_bases
        if rpkm < rpkm_threshold:
          pi, prr, pp_count, gb_count = 'gene not expressed', 'gene not expressed', 'gene not expressed', 'gene not expressed'

        if e - s < gene_length:
          pi, prr, pp_count, gb_count='gene too short', 'gene too short', 'gene too short', 'gene too short'
        else:
          pi, pp_count, gb_count = pausing_index(c,s, e, strand, bw, tss_up, tss_down)
          # prr = elongation_index(c,s, e, strand, bw)
          if pi == 0:
            prr = 0
          else:
            prr = math.log2(1/pi)
      except:
        continue
      gene_rpkm_dic[gene_name] = rpkm
      gene_baseCount_dic[gene_name] = base_count
      gene_rpm_dic[gene_name] = rpm
      gene_pi_dic[gene_name] = pi
      gene_ei_dic[gene_name] = prr
      gene_pp_count_dic[gene_name] = pp_count
      gene_gb_count_dic[gene_name] = gb_count

      try:
        arr = bw.values(c, s, e)
        region_list = search_truncated_region(arr)
        if len(region_list) > 0:
          for region in region_list:
            if strand == '+':
              trunc_gene_left = s+ region[0] -100
              trunc_gene_right = s + region[0] + region[1]
            else:
              trunc_gene_left = s+ region[0]
              trunc_gene_right = s+ region[0] + region[1] + 100

            density = bw.stats(c, trunc_gene_left, trunc_gene_right)[0]
            base_count = int( density * (trunc_gene_right - trunc_gene_left) )
            trunc_gene_list.append([c, str(trunc_gene_left), str(trunc_gene_right), strand, str(base_count)])
      except:
        continue

    # dic2json(gene_rpkm_dic, rpkm_json_output)
  return gene_rpkm_dic, gene_baseCount_dic, gene_rpm_dic, \
  gene_pi_dic, gene_ei_dic, \
  gene_pp_count_dic, gene_gb_count_dic, trunc_gene_list

def coding_vs_linc(protein_data, linc_data, output_root):
  protein_num, linc_num = len(protein_data), len(linc_data)
  big_labels = [f'Coding\n({protein_num})', f'LncRNA\n({linc_num})']
  expressed_protein_num, expressed_linc_num = len(protein_data[protein_data>0]), len(linc_data[linc_data>0])
  expressed_protein_ratio = np.round( expressed_protein_num/protein_num, 2) * 100
  expressed_linc_ratio = np.round( expressed_linc_num/linc_num, 2) * 100
  big_sizes = [protein_num, linc_num]

  small_labels = [f'{expressed_protein_ratio}% Protein-coding genes\nExpressed',
    f'{100-expressed_protein_ratio}% Protein-coding genes\nNot Expressed',
    f'{expressed_linc_ratio}% LncRNA genes\nExpressed',
    f'{100-expressed_linc_ratio}% LncRNA genes\nNot Expressed'
  ]
  small_sizes = [expressed_protein_num, protein_num-expressed_protein_num, expressed_linc_num, linc_num - expressed_linc_num]

  # bigger
  expressed_pie_data = { 'big_sizes': big_sizes,
  'big_labels': big_labels,
  'small_sizes': small_sizes,
  'small_labels': small_labels}
  expressed_pie(expressed_pie_data, output_root)
  # with open('expressed_pie_data.pickle', 'wb') as f:
  #   pickle.dump(expressed_pie_data, f)

  expressed_dist_data = {'protein_data': protein_data, 'linc_data': linc_data }
  expressed_distribution(expressed_dist_data, output_root)
  # with open('expressed_dist_data.pickle', 'wb') as f:
  #   pickle.dump(expressed_dist_data, f)

# count, RPKM, PI, EI, Exon/intron density
def main(gtf, forward_bw, reverse_bw, output_root, tss_up=150, tss_down=150, gene_length=2000, rpkm_threshold=0.1):
  # 这里 多搞两个 参数用 tss_up, tss_down 去定义 pp区域，
  # tes的部分暂时不提，就用tss_right 到 tes 定义gb区域。
  total_bases = get_total_bases(forward_bw, reverse_bw)
  gene_df = get_gene_df( gtf )
  filter_gene_df = gene_df[ gene_df['genetype'].isin(['protein_coding', 'lincRNA', 'lncRNA']) ]
  gene_range_dic = gene_df2dic(filter_gene_df)
  # count, RPKM, promoter, geneBody count, PI, EI

  gene_rpkm_dic, gene_baseCount_dic, gene_rpm_dic, \
  gene_pi_dic, gene_ei_dic, \
  gene_pp_count_dic, gene_gb_count_dic, \
  trunk_gene_list = get_attrs(gene_range_dic, forward_bw, reverse_bw, total_bases, tss_up, tss_down, gene_length, rpkm_threshold)
  with open(output_root+'csv/trunc_genes_count.csv', 'w') as f:
    f.write('\t'.join(['trunc_gene_name', 'chrom', 'start', 'end', 'strand', 'count']) + '\n' )
    for x in trunk_gene_list:
      gene_name = x[0] + ':' +  str(x[1]) + '-' + str(x[2])
      x.insert(0, gene_name)
      f.write( '\t'.join( x ) + '\n')

  # print( list(gene_rpkm_dic.items())[:5] )

  attr_list = ['baseCount', 'rpkm', 'rpm', 'pp_count', 'gb_count', 'pi', 'ei']
  filter_gene_type = set(filter_gene_df['genetype'])
  for gene_type in filter_gene_type:
    gene_list = filter_gene_df[filter_gene_df['genetype'] == gene_type]['gene_name']
    for attr in attr_list:
      # type_attr_dic = {g: globals()['gene_' + attr +'_dic'][g] for g in gene_list}
      tmp_dic = eval('gene_' + attr +'_dic')
      tmp_list = list( tmp_dic.keys() )
      type_attr_dic = {g: tmp_dic[g] for g in gene_list if g in tmp_list}
      pd.DataFrame( {attr: pd.Series(type_attr_dic)} ).to_csv(output_root+'csv/' + gene_type + '_'+ attr + '.csv')


  gene_attrs_df = pd.concat({'rpkm':pd.Series(gene_rpkm_dic),
    'count': pd.Series(gene_baseCount_dic),
    'rpm': pd.Series(gene_rpm_dic),
    'pp_count': pd.Series(gene_pp_count_dic),
    'gb_count': pd.Series(gene_gb_count_dic),
    'pi': pd.Series(gene_pi_dic),
    'ei': pd.Series(gene_ei_dic)}, axis=1)

  gene_attrs_df.to_csv(output_root+'csv/%s_feature_attrs.csv'%gene_type)

  chr_list = list( set( filter_gene_df['chrom'] ) )

  sort_chr_list = [chrom for chrom in chr_list if re.match( r'chr\d+', chrom)]
  sort_chr_list.sort(key=lambda arr: (arr[:3], int(arr[3:])))
  # if 'chrM' in chr_list:  sort_chr_list.append('chrM')
  # if 'chrMT' in chr_list:  sort_chr_list.append('chrMT')
  if 'chrX' in chr_list:  sort_chr_list.append('chrX')
  if 'chrY' in chr_list:  sort_chr_list.append('chrY')
  rpkm_list = []
  for chrom in sort_chr_list:
    gene_list = filter_gene_df[ filter_gene_df['chrom'] == chrom ]['gene_name']

    cur_rpkm_list = []
    for gene in gene_list:
      try:
        cur_rpkm_list.append( np.log10(gene_rpkm_dic[gene]) )
      except:
        continue
    rpkm_list.append( cur_rpkm_list )
  chr_rpkm_data = {'rpkm_list': rpkm_list, 'sort_chr_list': sort_chr_list}
  chr_rpkm_boxplot( chr_rpkm_data, output_root )
  # with open('chr_rpkm_data.pickle', 'wb') as f:
  #   pickle.dump(chr_rpkm_data, f)

  # protein vs linc
  if ('protein_coding' in filter_gene_type) and ('lincRNA' in filter_gene_type):
    protein_data = pd.read_csv(output_root + 'csv/protein_coding_rpkm.csv').rpkm
    linc_data = pd.read_csv( output_root + 'csv/lincRNA_rpkm.csv').rpkm
    coding_vs_linc(protein_data, linc_data, output_root)
  if ('protein_coding' in filter_gene_type) and ('lncRNA' in filter_gene_type):
    protein_data = pd.read_csv(output_root + 'csv/protein_coding_rpkm.csv').rpkm
    linc_data = pd.read_csv( output_root + 'csv/lncRNA_rpkm.csv').rpkm
    coding_vs_linc(protein_data, linc_data, output_root)


  # Exon/Intron
  protein_exon_range_dic = get_gene_merge_exon_dic(gtf)
  forward_bw_signal = pyBigWig.open(  forward_bw )
  reverse_bw_signal = pyBigWig.open(  reverse_bw )
  exon_intron_ratio_dic = {}
  for gene_name, gr_dic in protein_exon_range_dic.items():
    chrom, strand, exon_range_list = gr_dic['chrom'], gr_dic['strand'], gr_dic['exon']
    try:
      gene_range = gene_range_dic[strand][gene_name]
    except:
      # print('no gene', gene_name)
      os.sys.exit(0)
      continue
    gene_start, gene_end = gene_range[1], gene_range[2]
    if strand == '+':
      bw = forward_bw_signal
    else:
      bw = reverse_bw_signal
    gene_length = gene_end - gene_start
    try:
      gene_total = bw.stats(chrom, gene_start, gene_end)[0] * gene_length
    except:
      # print(gene_name, chrom, gene_start, gene_end)
      continue
    exon_length, exon_total = 0, 0
    for exon_range in exon_range_list:
      cur_length = (exon_range[1] - exon_range[0])
      exon_length+=cur_length
      try:
        exon_total +=( cur_length* bw.stats(chrom, exon_range[0], exon_range[1])[0] )
      except:
        continue
    intron_total = gene_total - exon_total
    intron_length = gene_length - exon_length

    if intron_total <= 0:
      # print( "warning: intron density <=0")
      exon_density, intron_density,  ratio = exon_total/exon_length, 0, 0
    else:
      exon_density, intron_density = exon_total/exon_length, intron_total/intron_length
      ratio = exon_density/intron_density
    exon_intron_ratio_dic[gene_name] = {'exon': exon_density, 'intron': intron_density, 'ratio': ratio}

  exon_intron_df = pd.DataFrame(exon_intron_ratio_dic)
  exon_intron_df = exon_intron_df.T
  exon_intron_df.to_csv(output_root+'csv/exon_intron_ratio.csv')

  filter_exon_intron_df = exon_intron_df[(exon_intron_df['exon'] > 0) & (exon_intron_df['intron'] >0) ]

  '''
  slope, intercept, r_value, p_value, std_err = stats.linregress(filter_exon_intron_df['exon'],filter_exon_intron_df['intron'])
  text= 'y = {:4.2e} x + {:4.2e}; \nR^2= {:2.2f}; p= {:2.2f}'.format(slope, intercept, r_value*r_value, p_value)
  scatterplot( filter_exon_intron_df['exon'], filter_exon_intron_df['intron'], 'Exon/Intron ratio', 'Exon', 'Intron', text, output_root+'imgs/exon_intron_ratio.png' )
  '''

  # remove 两端的异常值
  exon_min, exon_max = filter_exon_intron_df['exon'].quantile( [0.1, 0.9] )
  intron_min, intron_max = filter_exon_intron_df['intron'].quantile( [0.1, 0.9] )
  filter_exon_intron_df1 = filter_exon_intron_df[(filter_exon_intron_df['exon'] > exon_min) & (filter_exon_intron_df['intron'] >intron_min) & (filter_exon_intron_df['exon'] < exon_max) & (filter_exon_intron_df['intron'] <intron_max) ]
  # 计算拟合值
  slope, intercept, r_value, p_value, std_err = stats.linregress( filter_exon_intron_df1['exon'], filter_exon_intron_df1['intron'] )
  text= 'y = {:4.2e} x + {:4.2e}; \nR^2= {:2.2f}; p= {:2.2e}'.format(slope, intercept, r_value*r_value, p_value)

  tmp_dir = output_root + 'csv/preprocess_report.csv'
  if os.path.exists(tmp_dir):
    with open(tmp_dir, encoding="utf-8",mode="a") as file:
        file.write('mRNA_contamination' + ',' + str(slope) + '\n')
  else:
    with open(output_root + 'csv/preprocess_report.csv', 'w') as file:
      file.write( 'mRNA_contamination' + ',' + str(slope) + '\n' )

  exon_intron_data = { 'filter_exon_intron_df1': filter_exon_intron_df1, 'text': text}
  exon_intron_scatter( exon_intron_data, output_root )
  # with open('exon_intron_data.pickle', 'wb') as f:
  #   pickle.dump(exon_intron_data, f)


  # proximal pause site
  pp_site_file = open(output_root + 'bed/proximal_pause_sites.bed', 'w')
  if ('protein_coding' in filter_gene_type):
    protein_data = pd.read_csv(output_root + 'csv/protein_coding_pi.csv', index_col=0)
    filter_genes = protein_data[pd.to_numeric( protein_data.pi, errors='coerce' ) > 2].index
    site_index_list = []
    for i, row in filter_gene_df[filter_gene_df['gene_name'].isin(filter_genes)].iterrows():
      chr = row['chrom']
      gene_name = row['gene_name']
      if row['strand'] == '+':
        tss = row['start']
        # density_list = [forward_bw_signal.stats(chr, tss+j*50, tss+(j+1)*50 )[0] for j in range(7)]
        signals = forward_bw_signal.values(chr, tss, tss+350 )
        signal_median = np.median( signals )
        if signal_median < 0.5: continue
        max_idx = np.argmax( signals )
        if signals[max_idx]/signal_median < 5: continue
        ln = chr + '\t' + str( tss + max_idx ) + '\t' +str( tss + max_idx+1 ) + '\t' + gene_name + '\t' + str(max_idx ) + '\n'
      else:
        tss = row['end']
        signals = forward_bw_signal.values(chr, tss-350, tss )
        signal_median = np.median( signals )
        if signal_median < 0.5: continue
        max_idx = np.argmax( signals )
        if signals[max_idx]/signal_median < 5: continue
        # density_list = [reverse_bw_signal.stats(chr, tss-(j+1)*50, tss-j*50 )[0] for j in range(7)]
        ln = chr + '\t' + str( tss - (350-max_idx+1) ) + '\t' + str( tss -  (350-max_idx) ) + '\t' + gene_name + '\t' +  str(max_idx ) + '\n'
      # print(ln)
      pp_site_file.write(ln)
      site_index_list.append( max_idx )
  # print( site_index_list )
  pp_site_file.close()
  if len(site_index_list) > 0:
    proximal_pausing_sites_distribution(site_index_list, output_root)
  # with open('site_index_list.pickle', 'wb') as f:
  #   pickle.dump(site_index_list, f)


if __name__ == '__main__':
  fire.Fire( main )